# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json

import pytest
import pytest_httpx

import polarion_rest_api_client as polarion_api
from tests.conftest import (
    TEST_STEPS_CREATED_RESPONSE,
    TEST_STEPS_NEXT_RESPONSE,
    TEST_STEPS_NO_NEXT_RESPONSE,
    TEST_STEPS_PATCH_REQUEST,
    TEST_STEPS_POST_REQUEST,
)


@pytest.mark.parametrize(
    ("revision", "query"),
    [
        (
            None,
            {
                "page[size]": "100",
                "fields[teststeps]": "@all",
            },
        ),
        (
            "12345",
            {
                "page[size]": "100",
                "fields[teststeps]": "@all",
                "revision": "12345",
            },
        ),
    ],
    ids=["no_revision", "with_revision"],
)
def test_get_test_steps_multi_page(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
    revision: str | None,
    query: dict,
):
    with open(TEST_STEPS_NEXT_RESPONSE, encoding="utf8") as f:
        content = json.load(f)
        httpx_mock.add_response(json=content)
        httpx_mock.add_response(json=content)

    with open(TEST_STEPS_NO_NEXT_RESPONSE, encoding="utf8") as f:
        httpx_mock.add_response(json=json.load(f))

    test_steps: list[polarion_api.TestStep] = (
        client.work_items.test_steps.get_all(
            "123", fields={"teststeps": "@all"}, revision=revision
        )
    )

    reqs = httpx_mock.get_requests()

    assert len(reqs) == 3
    assert reqs[0].method == "GET"
    query["page[number]"] = "1"
    assert dict(reqs[0].url.params) == query
    query["page[number]"] = "2"
    assert dict(reqs[1].url.params) == query
    query["page[number]"] = "3"
    assert dict(reqs[2].url.params) == query
    assert len(test_steps) == 3
    assert all(ts.step_index == 1 for ts in test_steps)
    assert all(ts.revision == "1234" for ts in test_steps)
    assert all(ts.work_item_id == "MyWorkItemId" for ts in test_steps)
    assert all(
        ts.step_columns
        == {"string": polarion_api.TextContent("text/html", "My text value")}
        for ts in test_steps
    )


def test_create_test_steps(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_STEPS_CREATED_RESPONSE, encoding="utf8") as f:
        httpx_mock.add_response(201, json=json.load(f))

    with open(TEST_STEPS_POST_REQUEST, encoding="utf8") as f:
        expected_req = json.load(f)

    work_item_id = "asdfg"

    ts_1 = polarion_api.TestStep(
        work_item_id,
        step_columns={"string": polarion_api.HtmlContent("My text value")},
    )
    ts_2 = polarion_api.TestStep(
        work_item_id,
        step_columns={"string": polarion_api.HtmlContent("My text value")},
    )

    client.work_items.test_steps.create([ts_1, ts_2])

    reqs = httpx_mock.get_requests()
    assert len(reqs) == 1
    req_data = json.loads(reqs[0].content.decode("utf-8"))
    assert req_data == expected_req
    assert reqs[0].url.path.endswith(f"/workitems/{work_item_id}/teststeps")
    assert ts_1.step_index == 0
    assert ts_2.step_index == 1


def test_update_test_step(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    httpx_mock.add_response(204)
    httpx_mock.add_response(204)
    with open(TEST_STEPS_PATCH_REQUEST, encoding="utf8") as f:
        expected_req = json.load(f)

    work_item_1_id = "asdfg"
    work_item_2_id = "qwertz"
    ts_1 = polarion_api.TestStep(
        work_item_1_id,
        1,
        step_columns={"string": polarion_api.HtmlContent("My text value")},
    )
    ts_2 = polarion_api.TestStep(
        work_item_1_id,
        2,
        step_columns={"string": polarion_api.HtmlContent("My text value")},
    )
    ts_3 = polarion_api.TestStep(
        work_item_2_id,
        1,
        step_columns={"string": polarion_api.HtmlContent("My text value")},
    )

    client.work_items.test_steps.update([ts_1, ts_2, ts_3])

    reqs = httpx_mock.get_requests()
    assert len(reqs) == 2
    req_data = json.loads(reqs[0].content.decode("utf-8"))
    assert req_data == expected_req
    assert reqs[0].url.path.endswith(f"/workitems/{work_item_1_id}/teststeps")
    assert reqs[1].url.path.endswith(f"/workitems/{work_item_2_id}/teststeps")



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.collections_list_post_request_data_item_relationships_documents import (
        CollectionsListPostRequestDataItemRelationshipsDocuments,
    )
    from ..models.collections_list_post_request_data_item_relationships_rich_pages import (
        CollectionsListPostRequestDataItemRelationshipsRichPages,
    )
    from ..models.collections_list_post_request_data_item_relationships_upstream_collections import (
        CollectionsListPostRequestDataItemRelationshipsUpstreamCollections,
    )


T = TypeVar("T", bound="CollectionsListPostRequestDataItemRelationships")


@_attrs_define
class CollectionsListPostRequestDataItemRelationships:
    """
    Attributes:
        documents (Union[Unset, CollectionsListPostRequestDataItemRelationshipsDocuments]):
        rich_pages (Union[Unset, CollectionsListPostRequestDataItemRelationshipsRichPages]):
        upstream_collections (Union[Unset, CollectionsListPostRequestDataItemRelationshipsUpstreamCollections]):
    """

    documents: Union[
        Unset, "CollectionsListPostRequestDataItemRelationshipsDocuments"
    ] = UNSET
    rich_pages: Union[
        Unset, "CollectionsListPostRequestDataItemRelationshipsRichPages"
    ] = UNSET
    upstream_collections: Union[
        Unset,
        "CollectionsListPostRequestDataItemRelationshipsUpstreamCollections",
    ] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        documents: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.documents, Unset):
            documents = self.documents.to_dict()

        rich_pages: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.rich_pages, Unset):
            rich_pages = self.rich_pages.to_dict()

        upstream_collections: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.upstream_collections, Unset):
            upstream_collections = self.upstream_collections.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if documents is not UNSET:
            field_dict["documents"] = documents
        if rich_pages is not UNSET:
            field_dict["richPages"] = rich_pages
        if upstream_collections is not UNSET:
            field_dict["upstreamCollections"] = upstream_collections

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.collections_list_post_request_data_item_relationships_documents import (
            CollectionsListPostRequestDataItemRelationshipsDocuments,
        )
        from ..models.collections_list_post_request_data_item_relationships_rich_pages import (
            CollectionsListPostRequestDataItemRelationshipsRichPages,
        )
        from ..models.collections_list_post_request_data_item_relationships_upstream_collections import (
            CollectionsListPostRequestDataItemRelationshipsUpstreamCollections,
        )

        d = dict(src_dict)
        _documents = d.pop("documents", UNSET)
        documents: Union[
            Unset, CollectionsListPostRequestDataItemRelationshipsDocuments
        ]
        if isinstance(_documents, Unset):
            documents = UNSET
        else:
            documents = CollectionsListPostRequestDataItemRelationshipsDocuments.from_dict(
                _documents
            )

        _rich_pages = d.pop("richPages", UNSET)
        rich_pages: Union[
            Unset, CollectionsListPostRequestDataItemRelationshipsRichPages
        ]
        if isinstance(_rich_pages, Unset):
            rich_pages = UNSET
        else:
            rich_pages = CollectionsListPostRequestDataItemRelationshipsRichPages.from_dict(
                _rich_pages
            )

        _upstream_collections = d.pop("upstreamCollections", UNSET)
        upstream_collections: Union[
            Unset,
            CollectionsListPostRequestDataItemRelationshipsUpstreamCollections,
        ]
        if isinstance(_upstream_collections, Unset):
            upstream_collections = UNSET
        else:
            upstream_collections = CollectionsListPostRequestDataItemRelationshipsUpstreamCollections.from_dict(
                _upstream_collections
            )

        collections_list_post_request_data_item_relationships_obj = cls(
            documents=documents,
            rich_pages=rich_pages,
            upstream_collections=upstream_collections,
        )

        collections_list_post_request_data_item_relationships_obj.additional_properties = d
        return collections_list_post_request_data_item_relationships_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import itertools
import typing as t

from polarion_rest_api_client import data_models as dm
from polarion_rest_api_client.open_api_client import models as api_models
from polarion_rest_api_client.open_api_client import types as oa_types
from polarion_rest_api_client.open_api_client.api.test_steps import (
    delete_test_steps,
    get_test_steps,
    patch_test_steps,
    post_test_steps,
)

from . import base_classes as bc


class TestSteps(
    bc.MultiGetClient[dm.TestStep],
    bc.UpdateClient[dm.TestStep],
    bc.DeleteClient[dm.TestStep],
    bc.CreateClient[dm.TestStep],
):
    def _update(self, to_update: list[dm.TestStep]) -> None:
        response = patch_test_steps.sync_detailed(
            self._project_id,
            to_update[0].work_item_id,
            client=self._client.client,
            body=self._build_patch_request_data(to_update),
        )
        self._raise_on_error(response)

    async def _async_update(self, to_update: list[dm.TestStep]) -> None:
        response = await patch_test_steps.asyncio_detailed(
            self._project_id,
            to_update[0].work_item_id,
            client=self._client.client,
            body=self._build_patch_request_data(to_update),
        )
        self._raise_on_error(response)

    def get_multi(  # type: ignore[override]
        self,
        work_item_id: str,
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
        revision: str | None = None,
    ) -> tuple[list[dm.TestStep], bool]:
        if fields is None:
            fields = self._client.default_fields.teststeps

        sparse_fields = self._build_sparse_fields(fields)
        response = get_test_steps.sync_detailed(
            self._project_id,
            work_item_id,
            client=self._client.client,
            fields=sparse_fields,
            pagenumber=page_number,
            pagesize=page_size,
            revision=revision or oa_types.UNSET,
        )
        return self._parse_get_response(response)

    async def async_get_multi(  # type: ignore[override]
        self,
        work_item_id: str,
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
        revision: str | None = None,
    ) -> tuple[list[dm.TestStep], bool]:
        if fields is None:
            fields = self._client.default_fields.teststeps

        sparse_fields = self._build_sparse_fields(fields)
        response = await get_test_steps.asyncio_detailed(
            self._project_id,
            work_item_id,
            client=self._client.client,
            fields=sparse_fields,
            pagenumber=page_number,
            pagesize=page_size,
            revision=revision or oa_types.UNSET,
        )
        return self._parse_get_response(response)

    def _parse_get_response(
        self, response: oa_types.Response
    ) -> tuple[list[dm.TestStep], bool]:
        self._raise_on_error(response)
        parsed_response = response.parsed
        assert isinstance(parsed_response, api_models.TeststepsListGetResponse)
        test_steps = [
            dm.TestStep(
                work_item_id=data.id.split("/")[1],
                step_index=int(data.id.split("/")[2]),
                revision=data.revision if data.revision else None,
                step_columns=self._make_step_columns(data.attributes),
            )
            for data in parsed_response.data or []
            if isinstance(data.id, str)
        ]
        next_page = isinstance(
            parsed_response.links, api_models.TeststepsListGetResponseLinks
        ) and bool(parsed_response.links.next_)
        return test_steps, next_page

    def _make_step_columns(
        self,
        attributes: (
            api_models.TeststepsListGetResponseDataItemAttributes
            | oa_types.Unset
        ),
    ) -> dict[str, dm.TextContent]:
        if isinstance(attributes, oa_types.Unset):
            return {}

        keys = attributes.keys or []
        values = [
            dm.TextContent(
                val.type_.value if val.type_ else "text/plain",
                val.value or None,
            )
            for val in attributes.values or []
        ]
        assert len(keys) == len(values)
        return dict(zip(keys, values, strict=False))

    def _pre_batching_grouping(
        self, items: list[dm.TestStep]
    ) -> t.Generator[list[dm.TestStep], None, None]:
        for _, group in itertools.groupby(items, lambda x: x.work_item_id):
            yield list(group)

    def _create(self, items: list[dm.TestStep]) -> None:
        response = post_test_steps.sync_detailed(
            self._project_id,
            items[0].work_item_id,
            client=self._client.client,
            body=self._build_post_request_data(items),
        )
        self._process_post_response(items, response)

    async def _async_create(self, items: list[dm.TestStep]) -> None:
        response = await post_test_steps.asyncio_detailed(
            self._project_id,
            items[0].work_item_id,
            client=self._client.client,
            body=self._build_post_request_data(items),
        )
        self._process_post_response(items, response)

    def _process_post_response(
        self,
        items: list[dm.TestStep],
        response: oa_types.Response,
    ) -> None:
        self._raise_on_error(response)
        parsed_response = response.parsed
        assert isinstance(
            parsed_response, api_models.TeststepsListPostResponse
        )
        assert parsed_response.data
        for idx, response_item in enumerate(parsed_response.data):
            if response_item.id:
                items[idx].step_index = int(response_item.id.split("/")[-1])

    def _delete(self, items: list[dm.TestStep]) -> None:
        response = delete_test_steps.sync_detailed(
            self._project_id,
            items[0].work_item_id,
            client=self._client.client,
            body=self._build_delete_body(items),
        )
        self._raise_on_error(response)

    async def _async_delete(self, items: list[dm.TestStep]) -> None:
        response = await delete_test_steps.asyncio_detailed(
            self._project_id,
            items[0].work_item_id,
            client=self._client.client,
            body=self._build_delete_body(items),
        )
        self._raise_on_error(response)

    def _build_delete_body(
        self, items: list[dm.TestStep]
    ) -> api_models.TeststepsListDeleteRequest:
        return api_models.TeststepsListDeleteRequest(
            data=[
                api_models.TeststepsListDeleteRequestDataItem(
                    type_=api_models.TeststepsListDeleteRequestDataItemType.TESTSTEPS,
                    id=f"{self._project_id}/{step.work_item_id}/{step.step_index}",
                )
                for step in items
            ]
        )

    def _fill_test_step_attributes(
        self,
        attributes_type: type[
            api_models.TeststepsListPostRequestDataItemAttributes
            | api_models.TeststepsListPatchRequestDataItemAttributes
        ],
        test_step: dm.TestStep,
    ) -> (
        api_models.TeststepsListPostRequestDataItemAttributes
        | api_models.TeststepsListPatchRequestDataItemAttributes
    ):
        value_item_cls = getattr(
            api_models, f"{attributes_type.__name__}ValuesItem"
        )
        value_item_type = getattr(
            api_models, f"{attributes_type.__name__}ValuesItemType"
        )

        return attributes_type(
            keys=list(test_step.step_columns.keys()),
            values=[
                value_item_cls(
                    type_=(
                        value_item_type.TEXTHTML
                        if col.type == "text/html"
                        else value_item_type.TEXTPLAIN
                    ),
                    value=col.value,
                )
                for col in test_step.step_columns.values()
            ],
        )

    def _build_patch_request_data(
        self, items: list[dm.TestStep]
    ) -> api_models.TeststepsListPatchRequest:
        return api_models.TeststepsListPatchRequest(
            data=[
                api_models.TeststepsListPatchRequestDataItem(
                    type_=api_models.TeststepsListPatchRequestDataItemType.TESTSTEPS,
                    id=(
                        f"{self._project_id}/{step.work_item_id}/{step.step_index}"
                        if step.step_index
                        else oa_types.UNSET
                    ),
                    attributes=t.cast(
                        api_models.TeststepsListPatchRequestDataItemAttributes,
                        self._fill_test_step_attributes(
                            api_models.TeststepsListPatchRequestDataItemAttributes,
                            step,
                        ),
                    ),
                )
                for step in items
            ]
        )

    def _build_post_request_data(
        self, items: list[dm.TestStep]
    ) -> api_models.TeststepsListPostRequest:
        return api_models.TeststepsListPostRequest(
            data=[
                api_models.TeststepsListPostRequestDataItem(
                    type_=api_models.TeststepsListPostRequestDataItemType.TESTSTEPS,
                    attributes=t.cast(
                        api_models.TeststepsListPostRequestDataItemAttributes,
                        self._fill_test_step_attributes(
                            api_models.TeststepsListPostRequestDataItemAttributes,
                            step,
                        ),
                    ),
                )
                for step in items
            ]
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar(
    "T", bound="TeststepresultAttachmentsListPostRequestDataItemAttributes"
)


@_attrs_define
class TeststepresultAttachmentsListPostRequestDataItemAttributes:
    """
    Attributes:
        file_name (Union[Unset, str]):  Example: File Name.
        title (Union[Unset, str]):  Example: Title.
    """

    file_name: Union[Unset, str] = UNSET
    title: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        file_name = self.file_name

        title = self.title

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if file_name is not UNSET:
            field_dict["fileName"] = file_name
        if title is not UNSET:
            field_dict["title"] = title

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        file_name = d.pop("fileName", UNSET)

        title = d.pop("title", UNSET)

        teststepresult_attachments_list_post_request_data_item_attributes_obj = cls(
            file_name=file_name,
            title=title,
        )

        teststepresult_attachments_list_post_request_data_item_attributes_obj.additional_properties = d
        return teststepresult_attachments_list_post_request_data_item_attributes_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Implementations of WorkItemAttachment relates functions."""

import io

from polarion_rest_api_client import data_models as dm
from polarion_rest_api_client.open_api_client import models as api_models
from polarion_rest_api_client.open_api_client import types as oa_types
from polarion_rest_api_client.open_api_client.api.work_item_attachments import (  # pylint: disable=line-too-long
    delete_work_item_attachment,
    get_work_item_attachments,
    patch_work_item_attachment,
    post_work_item_attachments,
)

from . import base_classes as bc


class WorkItemAttachments(
    bc.UpdateClient[dm.WorkItemAttachment],
    bc.MultiGetClient[dm.WorkItemAttachment],
    bc.CreateClient[dm.WorkItemAttachment],
    bc.DeleteClient[dm.WorkItemAttachment],
):
    """A class to handle WorkItemAttachments."""

    _update_batch_size = 1
    _delete_batch_size = 1

    def _update(self, to_update: list[dm.WorkItemAttachment]) -> None:
        """Update the given work item attachment in Polarion."""
        item, multipart = self._prepare_patch_request(to_update)

        response = patch_work_item_attachment.sync_detailed(
            self._project_id,
            item.work_item_id,
            item.id,
            client=self._client.client,
            body=multipart,
        )
        self._raise_on_error(response)

    async def _async_update(
        self, to_update: list[dm.WorkItemAttachment]
    ) -> None:
        """Update the given work item attachment in Polarion."""
        item, multipart = self._prepare_patch_request(to_update)

        response = await patch_work_item_attachment.asyncio_detailed(
            self._project_id,
            item.work_item_id,
            item.id,
            client=self._client.client,
            body=multipart,
        )
        self._raise_on_error(response)

    def _prepare_patch_request(
        self, to_update: list[dm.WorkItemAttachment]
    ) -> tuple[
        dm.WorkItemAttachment, api_models.PatchWorkItemAttachmentsRequestBody
    ]:
        assert len(to_update) == 1, "Expected only one item"
        item = to_update[0]
        attributes = (
            api_models.WorkitemAttachmentsSinglePatchRequestDataAttributes()
        )
        if item.title:
            attributes.title = item.title
        multipart = api_models.PatchWorkItemAttachmentsRequestBody(
            resource=api_models.WorkitemAttachmentsSinglePatchRequest(
                data=api_models.WorkitemAttachmentsSinglePatchRequestData(
                    type_=api_models.WorkitemAttachmentsSinglePatchRequestDataType.WORKITEM_ATTACHMENTS,
                    # pylint: disable=line-too-long
                    id=f"{self._project_id}/{item.work_item_id}/{item.id}",
                    # pylint: disable=line-too-long
                    attributes=attributes,
                )
            )
        )
        if item.content_bytes:
            multipart.content = oa_types.File(
                io.BytesIO(item.content_bytes),
                item.file_name,
                item.mime_type,
            )
        return item, multipart

    def get_multi(  # type: ignore[override]
        self,
        work_item_id: str,
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
        revision: str | None = None,
    ) -> tuple[list[dm.WorkItemAttachment], bool]:
        """Return the attachments for a given work item on a defined page.

        In addition, a flag whether a next page is available is
        returned. Define a fields dictionary as described in the
        Polarion API documentation to get certain fields.
        """
        if fields is None:
            fields = self._client.default_fields.workitem_attachments

        sparse_fields = self._build_sparse_fields(fields)
        response = get_work_item_attachments.sync_detailed(
            self._project_id,
            work_item_id=work_item_id,
            client=self._client.client,
            fields=sparse_fields,
            pagesize=page_size,
            pagenumber=page_number,
            revision=revision or oa_types.UNSET,
        )

        return self._process_get_response(response, work_item_id)

    async def async_get_multi(  # type: ignore[override]
        self,
        work_item_id: str,
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
        revision: str | None = None,
    ) -> tuple[list[dm.WorkItemAttachment], bool]:
        """Return the attachments for a given work item on a defined page.

        In addition, a flag whether a next page is available is
        returned. Define a fields dictionary as described in the
        Polarion API documentation to get certain fields.
        """
        if fields is None:
            fields = self._client.default_fields.workitem_attachments

        sparse_fields = self._build_sparse_fields(fields)
        response = await get_work_item_attachments.asyncio_detailed(
            self._project_id,
            work_item_id=work_item_id,
            client=self._client.client,
            fields=sparse_fields,
            pagesize=page_size,
            pagenumber=page_number,
            revision=revision or oa_types.UNSET,
        )

        return self._process_get_response(response, work_item_id)

    def _process_get_response(
        self,
        response: oa_types.Response,
        work_item_id: str,
    ) -> tuple[list[dm.WorkItemAttachment], bool]:
        self._raise_on_error(response)
        parsed_response = response.parsed
        work_item_attachments: list[dm.WorkItemAttachment] = []
        next_page = False
        if (
            isinstance(
                parsed_response, api_models.WorkitemAttachmentsListGetResponse
            )
            and parsed_response.data
        ):
            for attachment in parsed_response.data:
                assert attachment.attributes
                assert isinstance(attachment.attributes.id, str)

                work_item_attachments.append(
                    dm.WorkItemAttachment(
                        work_item_id,
                        attachment.attributes.id,
                        self.unset_to_none(attachment.attributes.title),
                        file_name=self.unset_to_none(
                            attachment.attributes.file_name
                        ),
                    )
                )

            next_page = isinstance(
                parsed_response.links,
                api_models.WorkitemAttachmentsListGetResponseLinks,
            ) and bool(parsed_response.links.next_)
        return work_item_attachments, next_page

    def _create(self, items: list[dm.WorkItemAttachment]) -> None:
        """Create the given work item attachment in Polarion."""
        multipart = self._prepare_post_request(items)
        response = post_work_item_attachments.sync_detailed(
            self._project_id,
            items[0].work_item_id,
            client=self._client.client,
            body=multipart,
        )

        self._process_post_response(items, response)

    async def _async_create(self, items: list[dm.WorkItemAttachment]) -> None:
        """Create the given work item attachment in Polarion."""
        multipart = self._prepare_post_request(items)
        response = await post_work_item_attachments.asyncio_detailed(
            self._project_id,
            items[0].work_item_id,
            client=self._client.client,
            body=multipart,
        )

        self._process_post_response(items, response)

    def _process_post_response(
        self,
        items: list[dm.WorkItemAttachment],
        response: oa_types.Response,
    ) -> None:
        self._raise_on_error(response)
        parsed_response = response.parsed
        assert isinstance(
            parsed_response, api_models.WorkitemAttachmentsListPostResponse
        )
        assert parsed_response.data
        for counter, work_item_attachment_res in enumerate(
            parsed_response.data
        ):
            assert work_item_attachment_res.id
            items[counter].id = work_item_attachment_res.id.split("/")[-1]

    def _prepare_post_request(
        self, items: list[dm.WorkItemAttachment]
    ) -> api_models.PostWorkItemAttachmentsRequestBody:
        attachment_attributes = []
        attachment_files = []
        assert len(items), "No attachments were provided."
        assert all(
            [wia.work_item_id == items[0].work_item_id] for wia in items
        ), "All attachments must belong to the same WorkItem."
        for work_item_attachment in items:
            assert work_item_attachment.file_name, (
                "You have to define a FileName."
            )
            assert work_item_attachment.content_bytes, (
                "You have to provide content bytes."
            )
            assert work_item_attachment.mime_type, (
                "You have to provide a mime_type."
            )

            attributes = api_models.WorkitemAttachmentsListPostRequestDataItemAttributes(
                # pylint: disable=line-too-long
                file_name=work_item_attachment.file_name
            )
            if work_item_attachment.title:
                attributes.title = work_item_attachment.title

            attachment_attributes.append(
                api_models.WorkitemAttachmentsListPostRequestDataItem(
                    type_=api_models.WorkitemAttachmentsListPostRequestDataItemType.WORKITEM_ATTACHMENTS,
                    # pylint: disable=line-too-long
                    attributes=attributes,
                )
            )

            attachment_files.append(
                oa_types.File(
                    io.BytesIO(work_item_attachment.content_bytes),
                    work_item_attachment.file_name,
                    work_item_attachment.mime_type,
                )
            )
        return api_models.PostWorkItemAttachmentsRequestBody(
            resource=api_models.WorkitemAttachmentsListPostRequest(
                attachment_attributes
            ),
            files=attachment_files,
        )

    def _delete(self, items: list[dm.WorkItemAttachment]) -> None:
        assert len(items) == 1, "Expected only one item"
        item = items[0]
        response = delete_work_item_attachment.sync_detailed(
            self._project_id,
            item.work_item_id,
            item.id,
            client=self._client.client,
        )
        self._raise_on_error(response)

    async def _async_delete(self, items: list[dm.WorkItemAttachment]) -> None:
        assert len(items) == 1, "Expected only one item"
        item = items[0]
        response = await delete_work_item_attachment.asyncio_detailed(
            self._project_id,
            item.work_item_id,
            item.id,
            client=self._client.client,
        )
        self._raise_on_error(response)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import asyncio
import functools
import inspect
import json
import pathlib
import typing as t
from functools import wraps

import httpx
import pytest

import polarion_rest_api_client as polarion_api
from polarion_rest_api_client.clients import base_classes


@pytest.fixture(autouse=True)
def no_sleeps(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(
        base_classes.time, "sleep", lambda _delay: None, raising=True
    )

    async def _no_asyncio_sleep(*_: t.Any, **kwargs: t.Any) -> t.Any:
        return kwargs.get("result")

    monkeypatch.setattr(
        base_classes.asyncio, "sleep", _no_asyncio_sleep, raising=True
    )


def wrap_client(obj, is_async: bool):
    """Recursively wrap the client to select async methods if needed."""

    class ClientWrapper:
        def __getattr__(self, name):
            if is_async and hasattr(obj, f"async_{name}"):
                attr = getattr(obj, f"async_{name}")
            else:
                attr = getattr(obj, name)

            if inspect.iscoroutinefunction(attr) or (
                isinstance(attr, functools.partial)
                and inspect.iscoroutinefunction(attr.args[0])
            ):
                # Run coroutines sync and also catch those wrapped with retry_on_error
                @wraps(attr)
                def sync_wrapper(*args, **kwargs):
                    return asyncio.run(attr(*args, **kwargs))

                return sync_wrapper

            if callable(attr):
                return attr  # normal sync method

            # Wrap subclients, too
            return wrap_client(attr, is_async)

        def __setattr__(self, name: str, value: t.Any):
            setattr(obj, name, value)

    return ClientWrapper()


def pytest_generate_tests(metafunc):
    if "client" in metafunc.fixturenames:
        metafunc.parametrize("is_async", [False, True], ids=["sync", "async"])


@pytest.fixture
def client(is_async: bool):
    client = polarion_api.PolarionClient(
        polarion_api_endpoint="http://127.0.0.1/api",
        polarion_access_token="PAT123",
        batch_size=3,
    )
    client_base = client.generate_project_client(
        project_id="PROJ", delete_status="deleted"
    )
    return wrap_client(client_base, is_async)


@pytest.fixture(name="work_item")
def fixture_dummy_work_item():
    return polarion_api.WorkItem(
        title="Title",
        description=polarion_api.HtmlContent("My text value"),
        type="task",
        status="open",
        additional_attributes={"capella_uuid": "asdfg"},
    )


@pytest.fixture(name="work_item_patch")
def fixture_dummy_work_item_patch():
    return polarion_api.WorkItem(
        id="MyWorkItemId",
        title="Title",
        description=polarion_api.HtmlContent("My text value"),
        status="open",
        additional_attributes={
            "capella_uuid": "qwertz",
            "cfOwner": "NewOwner",
        },
    )


@pytest.fixture(name="work_item_attachment")
def fixture_dummy_work_item_attachment():
    with open(TEST_WIA_CREATED_RESPONSE, encoding="utf8") as f:
        attachment_content = f.read()

    return polarion_api.WorkItemAttachment(
        "MyWorkItemId",
        "Attachment",
        "Title",
        attachment_content.encode("utf-8"),
        "text/plain",
        "test.json",
    )


TEST_DATA_ROOT = pathlib.Path(__file__).parent / "data"
TEST_RESPONSES = TEST_DATA_ROOT / "mock_api_responses"
TEST_REQUESTS = TEST_DATA_ROOT / "expected_requests"
TEST_WIA_NEXT_PAGE_RESPONSE = (
    TEST_RESPONSES / "get_work_item_attachments_next_page.json"
)
TEST_WIA_NO_NEXT_PAGE_RESPONSE = (
    TEST_RESPONSES / "get_work_item_attachments_no_next_page.json"
)
TEST_WIA_CREATED_RESPONSE = (
    TEST_RESPONSES / "created_work_item_attachment.json"
)
TEST_WIA_MULTI_CREATED_RESPONSE = (
    TEST_RESPONSES / "created_work_item_attachments.json"
)
TEST_WIL_MULTI_POST_REQUEST = TEST_REQUESTS / "post_work_item_links.json"
TEST_WIL_DELETE2_REQUEST = TEST_REQUESTS / "delete_work_item_link_2.json"
TEST_WIL_DELETE_REQUEST = TEST_REQUESTS / "delete_work_item_links.json"
TEST_WIL_DELETED_REQUEST = TEST_REQUESTS / "delete_work_item_link.json"
TEST_WIL_POSTED_REQUEST = TEST_REQUESTS / "post_work_item_link.json"
TEST_WIL_CREATED_RESPONSE = TEST_RESPONSES / "created_work_item_links.json"
TEST_WIL_NEXT_PAGE_RESPONSE = (
    TEST_RESPONSES / "get_linked_work_items_next_page.json"
)
TEST_WIL_NO_NEXT_PAGE_RESPONSE = (
    TEST_RESPONSES / "get_linked_work_items_no_next_page.json"
)
TEST_WI_DELETE_REQUEST = TEST_REQUESTS / "delete_work_item.json"
TEST_WI_PATCH_STATUS_DELETED_REQUEST = (
    TEST_REQUESTS / "patch_work_item_status_deleted.json"
)
TEST_WI_PATCH_STATUS_REQUEST = TEST_REQUESTS / "patch_work_item_status.json"
TEST_WI_PATCH_TITLE_REQUEST = TEST_REQUESTS / "patch_work_item_title.json"
TEST_WI_PATCH_DESCRIPTION_REQUEST = (
    TEST_REQUESTS / "patch_work_item_description.json"
)
TEST_WI_PATCH_COMPLETELY_REQUEST = (
    TEST_REQUESTS / "patch_work_item_completely.json"
)
TEST_WI_PATCH_HYPERLINKS_REQUEST = (
    TEST_REQUESTS / "patch_work_item_hyperlinks.json"
)
TEST_WI_MULTI_POST_REQUEST = TEST_REQUESTS / "post_workitems.json"
TEST_WI_MULTI_POST_REQUEST_IN_DOC = TEST_REQUESTS / "post_workitem_in_doc.json"
TEST_WI_POST_REQUEST = TEST_REQUESTS / "post_workitem.json"
TEST_WI_NO_NEXT_PAGE_RESPONSE = TEST_RESPONSES / "workitems_no_next_page.json"
TEST_WI_CREATED_RESPONSE = TEST_RESPONSES / "created_work_items.json"
TEST_WI_ERROR_NEXT_PAGE_RESPONSE = (
    TEST_RESPONSES / "workitems_next_page_error.json"
)
TEST_WI_NEXT_PAGE_RESPONSE = TEST_RESPONSES / "workitems_next_page.json"
TEST_WI_SINGLE_RESPONSE = TEST_RESPONSES / "get_work_item.json"
TEST_WI_NOT_TRUNCATED_RESPONSE = (
    TEST_RESPONSES / "get_work_item_not_truncated.json"
)
TEST_DOCUMENT_RESPONSE = TEST_RESPONSES / "get_document.json"
TEST_DOCUMENT_POST_REQUEST = TEST_REQUESTS / "create_document.json"
TEST_DOCUMENT_PATCH_REQUEST = TEST_REQUESTS / "update_document.json"
TEST_DOCUMENT_PATCH_REQUEST2 = TEST_REQUESTS / "update_document_2.json"
TEST_ERROR_RESPONSE = TEST_RESPONSES / "error.json"
TEST_FAULTS_ERROR_RESPONSES = TEST_RESPONSES / "faulty_errors.json"
TEST_PROJECT_RESPONSE_JSON = TEST_RESPONSES / "project.json"
TEST_TRUN_PATCH_REQUEST = TEST_REQUESTS / "patch_test_run_partially.json"
TEST_TRUN_FULLY_PATCH_REQUEST = TEST_REQUESTS / "patch_test_run_fully.json"
TEST_TRUN_POST_REQUEST = TEST_REQUESTS / "post_test_run.json"
TEST_TREC_PATCH_REQUEST = TEST_REQUESTS / "patch_test_record.json"
TEST_TREC_PATCH_REQUEST_EX_BY = (
    TEST_REQUESTS / "patch_test_record_executed_by.json"
)
TEST_TREC_POST_REQUEST = TEST_REQUESTS / "post_test_records.json"
TEST_TREC_CREATED_RESPONSE = TEST_RESPONSES / "created_test_records.json"
TEST_TRUN_CREATED_RESPONSE = TEST_RESPONSES / "created_test_runs.json"
TEST_TREC_NEXT_RESPONSE = TEST_RESPONSES / "test_records_next_page.json"
TEST_TREC_NO_NEXT_RESPONSE = TEST_RESPONSES / "test_records_no_next_page.json"
TEST_TRUN_NEXT_RESPONSE = TEST_RESPONSES / "test_runs_next_page.json"
TEST_TRUN_NO_NEXT_RESPONSE = TEST_RESPONSES / "test_runs_no_next_page.json"

TEST_STEPS_CREATED_RESPONSE = TEST_RESPONSES / "created_test_steps.json"
TEST_STEPS_NEXT_RESPONSE = TEST_RESPONSES / "test_steps_next_page.json"
TEST_STEPS_NO_NEXT_RESPONSE = TEST_RESPONSES / "test_steps_no_next_page.json"
TEST_STEPS_PATCH_REQUEST = TEST_REQUESTS / "patch_test_steps.json"
TEST_STEPS_POST_REQUEST = TEST_REQUESTS / "post_test_steps.json"

TEST_TRUN_PARAM_NEXT_RES = TEST_RESPONSES / "get_test_run_parameter.json"
TEST_TRUN_PARAM_NO_NEXT_RES = (
    TEST_RESPONSES / "get_test_run_parameter_no_next.json"
)
TEST_TRUN_PARAM_CREATED_RES = (
    TEST_RESPONSES / "created_test_parameters_trun.json"
)
TEST_TRUN_PARAM_CREATED_REQ_1 = TEST_REQUESTS / "create_parameters_trun_1.json"
TEST_TRUN_PARAM_CREATED_REQ_2 = TEST_REQUESTS / "create_parameters_trun_2.json"
TEST_TRUN_PARAM_DELETE_REQ_1 = TEST_REQUESTS / "delete_parameters_trun_1.json"
TEST_TRUN_PARAM_DELETE_REQ_2 = TEST_REQUESTS / "delete_parameters_trun_2.json"


class CustomWorkItem(polarion_api.WorkItem):
    capella_uuid: str | None


def check_req(
    expected_path: str, req: httpx.Request, expected_body: pathlib.Path
):
    req_data = json.loads(req.content.decode("utf-8"))
    with open(expected_body, encoding="utf8") as f:
        expected_req = json.load(f)
    assert req_data == expected_req
    assert req.url.path.endswith(expected_path)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.errors import Errors
from ...models.sparse_fields import SparseFields
from ...models.usergroups_single_get_response import (
    UsergroupsSingleGetResponse,
)
from ...types import UNSET, Response, Unset


def _get_kwargs(
    group_id: str,
    *,
    fields: Union[Unset, "SparseFields"] = UNSET,
    include: Union[Unset, str] = UNSET,
    revision: Union[Unset, str] = UNSET,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    json_fields: Union[Unset, dict[str, Any]] = UNSET
    if not isinstance(fields, Unset):
        json_fields = fields.to_dict()
    if not isinstance(json_fields, Unset):
        params.update(json_fields)

    params["include"] = include

    params["revision"] = revision

    params = {
        k: v for k, v in params.items() if v is not UNSET and v is not None
    }

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": f"/usergroups/{group_id}",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Union[Errors, UsergroupsSingleGetResponse] | None:
    if response.status_code == 200:
        response_200 = UsergroupsSingleGetResponse.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = Errors.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = Errors.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = Errors.from_dict(response.json())

        return response_403
    if response.status_code == 404:
        response_404 = Errors.from_dict(response.json())

        return response_404
    if response.status_code == 406:
        response_406 = Errors.from_dict(response.json())

        return response_406
    if response.status_code == 500:
        response_500 = Errors.from_dict(response.json())

        return response_500
    if response.status_code == 503:
        response_503 = Errors.from_dict(response.json())

        return response_503
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Errors, UsergroupsSingleGetResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    group_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    fields: Union[Unset, "SparseFields"] = UNSET,
    include: Union[Unset, str] = UNSET,
    revision: Union[Unset, str] = UNSET,
) -> Response[Union[Errors, UsergroupsSingleGetResponse]]:
    """Returns the specified User Group.

    Args:
        group_id (str):
        fields (Union[Unset, SparseFields]):
        include (Union[Unset, str]):
        revision (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Errors, UsergroupsSingleGetResponse]]
    """

    kwargs = _get_kwargs(
        group_id=group_id,
        fields=fields,
        include=include,
        revision=revision,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    group_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    fields: Union[Unset, "SparseFields"] = UNSET,
    include: Union[Unset, str] = UNSET,
    revision: Union[Unset, str] = UNSET,
) -> Union[Errors, UsergroupsSingleGetResponse] | None:
    """Returns the specified User Group.

    Args:
        group_id (str):
        fields (Union[Unset, SparseFields]):
        include (Union[Unset, str]):
        revision (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Errors, UsergroupsSingleGetResponse]
    """

    return sync_detailed(
        group_id=group_id,
        client=client,
        fields=fields,
        include=include,
        revision=revision,
    ).parsed


async def asyncio_detailed(
    group_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    fields: Union[Unset, "SparseFields"] = UNSET,
    include: Union[Unset, str] = UNSET,
    revision: Union[Unset, str] = UNSET,
) -> Response[Union[Errors, UsergroupsSingleGetResponse]]:
    """Returns the specified User Group.

    Args:
        group_id (str):
        fields (Union[Unset, SparseFields]):
        include (Union[Unset, str]):
        revision (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Errors, UsergroupsSingleGetResponse]]
    """

    kwargs = _get_kwargs(
        group_id=group_id,
        fields=fields,
        include=include,
        revision=revision,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    group_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    fields: Union[Unset, "SparseFields"] = UNSET,
    include: Union[Unset, str] = UNSET,
    revision: Union[Unset, str] = UNSET,
) -> Union[Errors, UsergroupsSingleGetResponse] | None:
    """Returns the specified User Group.

    Args:
        group_id (str):
        fields (Union[Unset, SparseFields]):
        include (Union[Unset, str]):
        revision (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Errors, UsergroupsSingleGetResponse]
    """

    return (
        await asyncio_detailed(
            group_id=group_id,
            client=client,
            fields=fields,
            include=include,
            revision=revision,
        )
    ).parsed



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.testrecords_list_patch_request_data_item_type import (
    TestrecordsListPatchRequestDataItemType,
)
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.testrecords_list_patch_request_data_item_attributes import (
        TestrecordsListPatchRequestDataItemAttributes,
    )
    from ..models.testrecords_list_patch_request_data_item_relationships import (
        TestrecordsListPatchRequestDataItemRelationships,
    )


T = TypeVar("T", bound="TestrecordsListPatchRequestDataItem")


@_attrs_define
class TestrecordsListPatchRequestDataItem:
    """
    Attributes:
        type_ (Union[Unset, TestrecordsListPatchRequestDataItemType]):
        id (Union[Unset, str]):  Example: elibrary/MyTestRunId/MyProjectId/MyTestcaseId/0.
        attributes (Union[Unset, TestrecordsListPatchRequestDataItemAttributes]):
        relationships (Union[Unset, TestrecordsListPatchRequestDataItemRelationships]):
    """

    type_: Union[Unset, TestrecordsListPatchRequestDataItemType] = UNSET
    id: Union[Unset, str] = UNSET
    attributes: Union[
        Unset, "TestrecordsListPatchRequestDataItemAttributes"
    ] = UNSET
    relationships: Union[
        Unset, "TestrecordsListPatchRequestDataItemRelationships"
    ] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        type_: Union[Unset, str] = UNSET
        if not isinstance(self.type_, Unset):
            type_ = self.type_.value

        id = self.id

        attributes: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes.to_dict()

        relationships: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.relationships, Unset):
            relationships = self.relationships.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if type_ is not UNSET:
            field_dict["type"] = type_
        if id is not UNSET:
            field_dict["id"] = id
        if attributes is not UNSET:
            field_dict["attributes"] = attributes
        if relationships is not UNSET:
            field_dict["relationships"] = relationships

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.testrecords_list_patch_request_data_item_attributes import (
            TestrecordsListPatchRequestDataItemAttributes,
        )
        from ..models.testrecords_list_patch_request_data_item_relationships import (
            TestrecordsListPatchRequestDataItemRelationships,
        )

        d = dict(src_dict)
        _type_ = d.pop("type", UNSET)
        type_: Union[Unset, TestrecordsListPatchRequestDataItemType]
        if isinstance(_type_, Unset):
            type_ = UNSET
        else:
            type_ = TestrecordsListPatchRequestDataItemType(_type_)

        id = d.pop("id", UNSET)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, TestrecordsListPatchRequestDataItemAttributes]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = (
                TestrecordsListPatchRequestDataItemAttributes.from_dict(
                    _attributes
                )
            )

        _relationships = d.pop("relationships", UNSET)
        relationships: Union[
            Unset, TestrecordsListPatchRequestDataItemRelationships
        ]
        if isinstance(_relationships, Unset):
            relationships = UNSET
        else:
            relationships = (
                TestrecordsListPatchRequestDataItemRelationships.from_dict(
                    _relationships
                )
            )

        testrecords_list_patch_request_data_item_obj = cls(
            type_=type_,
            id=id,
            attributes=attributes,
            relationships=relationships,
        )

        testrecords_list_patch_request_data_item_obj.additional_properties = d
        return testrecords_list_patch_request_data_item_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import itertools
import typing as t

from polarion_rest_api_client import data_models as dm
from polarion_rest_api_client.open_api_client import models as api_models
from polarion_rest_api_client.open_api_client import types as oa_types
from polarion_rest_api_client.open_api_client.api.test_records import (
    delete_test_record,
    get_test_records,
    patch_test_record,
    post_test_records,
)

from . import base_classes as bc
from . import test_parameters

if t.TYPE_CHECKING:
    from polarion_rest_api_client import client as polarion_client

AttributesType = t.TypeVar(
    "AttributesType",
    bound=api_models.TestrecordsListPostRequestDataItemAttributes
    | api_models.TestrecordsSinglePatchRequestDataAttributes,
)


class TestRecords(
    bc.CreateClient[dm.TestRecord],
    bc.MultiGetClient[dm.TestRecord],
    bc.DeleteClient[dm.TestRecord],
    bc.UpdateClient[dm.TestRecord],
):
    _update_batch_size = 1

    def __init__(
        self, project_id: str, client: polarion_client.PolarionClient
    ):
        super().__init__(project_id, client)
        self.parameters = test_parameters.TestRecordParameters(
            project_id, client
        )

    def _update(self, to_update: list[dm.TestRecord]) -> None:
        assert len(to_update) == 1, "Expected only one item"
        item = to_update[0]
        response = patch_test_record.sync_detailed(
            self._project_id,
            item.test_run_id,
            item.work_item_project_id,
            item.work_item_id,
            str(item.iteration),
            client=self._client.client,
            body=self._build_patch_body(item),
        )
        self._raise_on_error(response)

    async def _async_update(self, to_update: list[dm.TestRecord]) -> None:
        assert len(to_update) == 1, "Expected only one item"
        item = to_update[0]
        response = await patch_test_record.asyncio_detailed(
            self._project_id,
            item.test_run_id,
            item.work_item_project_id,
            item.work_item_id,
            str(item.iteration),
            client=self._client.client,
            body=self._build_patch_body(item),
        )
        self._raise_on_error(response)

    def _build_patch_body(
        self, to_update: dm.TestRecord
    ) -> api_models.TestrecordsSinglePatchRequest:
        # pylint: disable=line-too-long
        return api_models.TestrecordsSinglePatchRequest(
            data=api_models.TestrecordsSinglePatchRequestData(
                type_=api_models.TestrecordsSinglePatchRequestDataType.TESTRECORDS,
                id=f"{self._project_id}/{to_update.test_run_id}/{to_update.work_item_project_id}/{to_update.work_item_id}/{to_update.iteration}",
                attributes=self._fill_test_record_attributes(
                    api_models.TestrecordsSinglePatchRequestDataAttributes,
                    to_update,
                ),
                relationships=api_models.TestrecordsSinglePatchRequestDataRelationships(
                    executed_by=api_models.TestrecordsSinglePatchRequestDataRelationshipsExecutedBy(
                        data=api_models.TestrecordsSinglePatchRequestDataRelationshipsExecutedByData(
                            type_=api_models.TestrecordsSinglePatchRequestDataRelationshipsExecutedByDataType.USERS,
                            id=to_update.executed_by,
                        )
                    ),
                )
                if to_update.executed_by is not None
                else oa_types.UNSET,
            )
        )
        # pylint: enable=line-too-long

    def get_multi(  # type: ignore[override]
        self,
        test_run_id: str,
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
    ) -> tuple[list[dm.TestRecord], bool]:
        if fields is None:
            fields = self._client.default_fields.testrecords

        sparse_fields = self._build_sparse_fields(fields)
        response = get_test_records.sync_detailed(
            self._project_id,
            test_run_id,
            client=self._client.client,
            fields=sparse_fields,
            pagenumber=page_number,
            pagesize=page_size,
        )
        return self._parse_get_response(response, test_run_id)

    async def async_get_multi(  # type: ignore[override]
        self,
        test_run_id: str,
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
    ) -> tuple[list[dm.TestRecord], bool]:
        if fields is None:
            fields = self._client.default_fields.testrecords

        sparse_fields = self._build_sparse_fields(fields)
        response = await get_test_records.asyncio_detailed(
            self._project_id,
            test_run_id,
            client=self._client.client,
            fields=sparse_fields,
            pagenumber=page_number,
            pagesize=page_size,
        )
        return self._parse_get_response(response, test_run_id)

    def _parse_get_response(
        self, response: oa_types.Response, test_run_id: str
    ) -> tuple[list[dm.TestRecord], bool]:
        self._raise_on_error(response)
        parsed_response = response.parsed
        assert isinstance(
            parsed_response, api_models.TestrecordsListGetResponse
        )
        test_records = []
        for data in parsed_response.data or []:
            assert isinstance(data.id, str)
            assert isinstance(
                data.attributes,
                api_models.TestrecordsListGetResponseDataItemAttributes,
            )
            _, _, project_id, work_item, iteration = data.id.split("/")
            executed_by = None
            if (
                data.relationships
                and (ex_by := data.relationships.executed_by)
                and ex_by.data
                and ex_by.data.id
            ):
                executed_by = ex_by.data.id
            test_records.append(
                dm.TestRecord(
                    test_run_id=test_run_id,
                    work_item_project_id=project_id,
                    work_item_id=work_item,
                    work_item_revision=self.unset_to_none(
                        data.attributes.test_case_revision
                    ),
                    iteration=int(iteration),
                    duration=(
                        data.attributes.duration
                        if not isinstance(
                            data.attributes.duration, oa_types.Unset
                        )
                        else -1
                    ),
                    result=self.unset_to_none(data.attributes.result),
                    comment=self._handle_text_content(data.attributes.comment),
                    executed=self.unset_to_none(data.attributes.executed),
                    executed_by=executed_by,
                    additional_attributes=data.additional_properties or {},
                )
            )
        next_page = isinstance(
            parsed_response.links,
            api_models.TestrecordsListGetResponseLinks,
        ) and bool(parsed_response.links.next_)
        return test_records, next_page

    def _pre_batching_grouping(
        self, items: list[dm.TestRecord]
    ) -> t.Generator[list[dm.TestRecord], None, None]:
        for _, group in itertools.groupby(items, lambda x: x.test_run_id):
            yield list(group)

    def _create(
        self,
        items: list[dm.TestRecord],
    ) -> None:
        """Create the given list of test records."""
        response = post_test_records.sync_detailed(
            self._project_id,
            items[0].test_run_id,
            client=self._client.client,
            # pylint: disable=line-too-long
            body=self._build_post_body(items),
            # pylint: enable=line-too-long
        )

        self._parse_post_response(items, response)

    async def _async_create(
        self,
        items: list[dm.TestRecord],
    ) -> None:
        """Create the given list of test records."""
        response = await post_test_records.asyncio_detailed(
            self._project_id,
            items[0].test_run_id,
            client=self._client.client,
            body=self._build_post_body(items),
        )
        self._parse_post_response(items, response)

    def _parse_post_response(
        self, items: list[dm.TestRecord], response: oa_types.Response
    ) -> None:
        self._raise_on_error(response)
        assert isinstance(
            response.parsed, api_models.TestrecordsListPostResponse
        )
        assert response.parsed.data
        counter = 0
        for response_item in response.parsed.data:
            if response_item.id:
                items[counter].iteration = int(response_item.id.split("/")[-1])
                counter += 1

    def _build_post_body(
        self, items: list[dm.TestRecord]
    ) -> api_models.TestrecordsListPostRequest:
        # pylint: disable=line-too-long
        return api_models.TestrecordsListPostRequest(
            [
                api_models.TestrecordsListPostRequestDataItem(
                    type_=api_models.TestrecordsListPostRequestDataItemType.TESTRECORDS,
                    attributes=self._fill_test_record_attributes(
                        api_models.TestrecordsListPostRequestDataItemAttributes,
                        test_record,
                    ),
                    relationships=api_models.TestrecordsListPostRequestDataItemRelationships(
                        test_case=api_models.TestrecordsListPostRequestDataItemRelationshipsTestCase(
                            data=api_models.TestrecordsListPostRequestDataItemRelationshipsTestCaseData(
                                type_=api_models.TestrecordsListPostRequestDataItemRelationshipsTestCaseDataType.WORKITEMS,
                                id=f"{test_record.work_item_project_id}/{test_record.work_item_id}",
                            )
                        ),
                        executed_by=api_models.TestrecordsListPostRequestDataItemRelationshipsExecutedBy(
                            data=api_models.TestrecordsListPostRequestDataItemRelationshipsExecutedByData(
                                type_=api_models.TestrecordsListPostRequestDataItemRelationshipsExecutedByDataType.USERS,
                                id=test_record.executed_by,
                            )
                        )
                        if test_record.executed_by is not None
                        else oa_types.UNSET,
                    ),
                )
                for test_record in items
            ]
        )
        # pylint: enable=line-too-long

    def _delete(self, items: list[dm.TestRecord]) -> None:
        assert len(items) == 1, "Expected only one item"
        item = items[0]
        response = delete_test_record.sync_detailed(
            self._project_id,
            item.test_run_id,
            item.work_item_project_id,
            item.work_item_id,
            str(item.iteration),
            client=self._client.client,
        )
        self._raise_on_error(response)

    async def _async_delete(self, items: list[dm.TestRecord]) -> None:
        assert len(items) == 1, "Expected only one item"
        item = items[0]
        response = await delete_test_record.asyncio_detailed(
            self._project_id,
            item.test_run_id,
            item.work_item_project_id,
            item.work_item_id,
            str(item.iteration),
            client=self._client.client,
        )
        self._raise_on_error(response)

    def _fill_test_record_attributes(
        self,
        attributes_type: type[AttributesType],
        test_record: dm.TestRecord,
    ) -> AttributesType:
        type_prefix = attributes_type.__name__
        attributes = attributes_type()
        if test_record.result:
            attributes.result = test_record.result
        if test_record.comment:
            attributes.comment = getattr(api_models, f"{type_prefix}Comment")()
            assert attributes.comment
            if test_record.comment.type:
                attributes.comment.type_ = getattr(
                    api_models, f"{type_prefix}CommentType"
                )(test_record.comment.type)
            if test_record.comment.value:
                attributes.comment.value = test_record.comment.value
        if test_record.duration != -1:
            attributes.duration = test_record.duration
        if test_record.work_item_revision:
            attributes.test_case_revision = test_record.work_item_revision
        if test_record.executed:
            attributes.executed = test_record.executed
        if test_record.additional_attributes:
            attributes.additional_properties = (
                test_record.additional_attributes
            )
        return t.cast(AttributesType, attributes)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.teststeps_list_post_request_data_item_attributes_values_item import (
        TeststepsListPostRequestDataItemAttributesValuesItem,
    )


T = TypeVar("T", bound="TeststepsListPostRequestDataItemAttributes")


@_attrs_define
class TeststepsListPostRequestDataItemAttributes:
    """
    Attributes:
        keys (Union[Unset, list[str]]):
        values (Union[Unset, list['TeststepsListPostRequestDataItemAttributesValuesItem']]):
    """

    keys: Union[Unset, list[str]] = UNSET
    values: Union[
        Unset, list["TeststepsListPostRequestDataItemAttributesValuesItem"]
    ] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        keys: Union[Unset, list[str]] = UNSET
        if not isinstance(self.keys, Unset):
            keys = self.keys

        values: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.values, Unset):
            values = []
            for values_item_data in self.values:
                values_item = values_item_data.to_dict()
                values.append(values_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if keys is not UNSET:
            field_dict["keys"] = keys
        if values is not UNSET:
            field_dict["values"] = values

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.teststeps_list_post_request_data_item_attributes_values_item import (
            TeststepsListPostRequestDataItemAttributesValuesItem,
        )

        d = dict(src_dict)
        keys = cast(list[str], d.pop("keys", UNSET))

        values = []
        _values = d.pop("values", UNSET)
        for values_item_data in _values or []:
            values_item = (
                TeststepsListPostRequestDataItemAttributesValuesItem.from_dict(
                    values_item_data
                )
            )

            values.append(values_item)

        teststeps_list_post_request_data_item_attributes_obj = cls(
            keys=keys,
            values=values,
        )

        teststeps_list_post_request_data_item_attributes_obj.additional_properties = d
        return teststeps_list_post_request_data_item_attributes_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.users_single_patch_request_data_relationships_global_roles import (
        UsersSinglePatchRequestDataRelationshipsGlobalRoles,
    )
    from ..models.users_single_patch_request_data_relationships_project_roles import (
        UsersSinglePatchRequestDataRelationshipsProjectRoles,
    )
    from ..models.users_single_patch_request_data_relationships_user_groups import (
        UsersSinglePatchRequestDataRelationshipsUserGroups,
    )


T = TypeVar("T", bound="UsersSinglePatchRequestDataRelationships")


@_attrs_define
class UsersSinglePatchRequestDataRelationships:
    """
    Attributes:
        global_roles (Union[Unset, UsersSinglePatchRequestDataRelationshipsGlobalRoles]):
        project_roles (Union[Unset, UsersSinglePatchRequestDataRelationshipsProjectRoles]):
        user_groups (Union[Unset, UsersSinglePatchRequestDataRelationshipsUserGroups]):
    """

    global_roles: Union[
        Unset, "UsersSinglePatchRequestDataRelationshipsGlobalRoles"
    ] = UNSET
    project_roles: Union[
        Unset, "UsersSinglePatchRequestDataRelationshipsProjectRoles"
    ] = UNSET
    user_groups: Union[
        Unset, "UsersSinglePatchRequestDataRelationshipsUserGroups"
    ] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        global_roles: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.global_roles, Unset):
            global_roles = self.global_roles.to_dict()

        project_roles: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.project_roles, Unset):
            project_roles = self.project_roles.to_dict()

        user_groups: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.user_groups, Unset):
            user_groups = self.user_groups.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if global_roles is not UNSET:
            field_dict["globalRoles"] = global_roles
        if project_roles is not UNSET:
            field_dict["projectRoles"] = project_roles
        if user_groups is not UNSET:
            field_dict["userGroups"] = user_groups

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.users_single_patch_request_data_relationships_global_roles import (
            UsersSinglePatchRequestDataRelationshipsGlobalRoles,
        )
        from ..models.users_single_patch_request_data_relationships_project_roles import (
            UsersSinglePatchRequestDataRelationshipsProjectRoles,
        )
        from ..models.users_single_patch_request_data_relationships_user_groups import (
            UsersSinglePatchRequestDataRelationshipsUserGroups,
        )

        d = dict(src_dict)
        _global_roles = d.pop("globalRoles", UNSET)
        global_roles: Union[
            Unset, UsersSinglePatchRequestDataRelationshipsGlobalRoles
        ]
        if isinstance(_global_roles, Unset):
            global_roles = UNSET
        else:
            global_roles = (
                UsersSinglePatchRequestDataRelationshipsGlobalRoles.from_dict(
                    _global_roles
                )
            )

        _project_roles = d.pop("projectRoles", UNSET)
        project_roles: Union[
            Unset, UsersSinglePatchRequestDataRelationshipsProjectRoles
        ]
        if isinstance(_project_roles, Unset):
            project_roles = UNSET
        else:
            project_roles = (
                UsersSinglePatchRequestDataRelationshipsProjectRoles.from_dict(
                    _project_roles
                )
            )

        _user_groups = d.pop("userGroups", UNSET)
        user_groups: Union[
            Unset, UsersSinglePatchRequestDataRelationshipsUserGroups
        ]
        if isinstance(_user_groups, Unset):
            user_groups = UNSET
        else:
            user_groups = (
                UsersSinglePatchRequestDataRelationshipsUserGroups.from_dict(
                    _user_groups
                )
            )

        users_single_patch_request_data_relationships_obj = cls(
            global_roles=global_roles,
            project_roles=project_roles,
            user_groups=user_groups,
        )

        users_single_patch_request_data_relationships_obj.additional_properties = d
        return users_single_patch_request_data_relationships_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Implementation of the documents client."""

import itertools
import logging
import typing as t
import urllib.parse

from polarion_rest_api_client import data_models as dm
from polarion_rest_api_client.open_api_client import models as api_models
from polarion_rest_api_client.open_api_client import types as oa_types
from polarion_rest_api_client.open_api_client.api.documents import (
    get_document,
    patch_document,
    post_documents,
)

from . import base_classes as bc

logger = logging.getLogger(__name__)


class Documents(
    bc.SingleGetClient,
    bc.CreateClient,
    bc.UpdateClient[dm.Document],
):
    """A client to work with documents in Polarion."""

    _update_batch_size = 1

    def get(
        self,
        space_id: str,
        document_name: str,
        fields: dict[str, str] | None = None,
        include: str | None | oa_types.Unset = None,
        revision: str | None | oa_types.Unset = None,
    ) -> dm.Document | None:
        """Return the document with the given document_name and space_id."""
        if fields is None:
            fields = self._client.default_fields.documents

        response = get_document.sync_detailed(
            self._project_id,
            urllib.parse.quote(space_id, safe="/", encoding=None, errors=None),
            urllib.parse.quote(
                document_name, safe="/", encoding=None, errors=None
            ),
            client=self._client.client,
            fields=self._build_sparse_fields(fields),
            include=self.none_to_unset(include),
            revision=self.none_to_unset(revision),
        )

        return self._parse_document_response(response)

    async def async_get(
        self,
        space_id: str,
        document_name: str,
        fields: dict[str, str] | None = None,
        include: str | None | oa_types.Unset = None,
        revision: str | None | oa_types.Unset = None,
    ) -> dm.Document | None:
        """Return the document with the given document_name and space_id."""
        if fields is None:
            fields = self._client.default_fields.documents

        response = await get_document.asyncio_detailed(
            self._project_id,
            urllib.parse.quote(space_id, safe="/", encoding=None, errors=None),
            urllib.parse.quote(
                document_name, safe="/", encoding=None, errors=None
            ),
            client=self._client.client,
            fields=self._build_sparse_fields(fields),
            include=self.none_to_unset(include),
            revision=self.none_to_unset(revision),
        )

        return self._parse_document_response(response)

    def _parse_document_response(
        self, response: oa_types.Response
    ) -> dm.Document | None:
        self._raise_on_error(response)
        document_response = response.parsed
        if (
            isinstance(
                document_response, api_models.DocumentsSingleGetResponse
            )
            and (data := document_response.data)
            and not getattr(data.meta, "errors", [])
        ):
            assert data.attributes
            attributes = data.attributes
            assert isinstance(data.id, str)
            home_page_content = self._handle_text_content(
                attributes.home_page_content
            )

            rendering_layouts = None
            if attributes.rendering_layouts:
                rendering_layouts = [
                    dm.RenderingLayout(
                        self.unset_to_none(layout.label),
                        self.unset_to_none(layout.layouter),
                        (
                            [p.to_dict() for p in layout.properties]
                            if layout.properties
                            else None
                        ),
                        self.unset_to_none(layout.type_),
                    )
                    for layout in attributes.rendering_layouts
                ]

            return dm.Document(
                id=data.id,
                module_folder=self.unset_to_none(attributes.module_folder),
                module_name=self.unset_to_none(attributes.module_name),
                type=self.unset_to_none(attributes.type_),
                status=self.unset_to_none(attributes.status),
                home_page_content=home_page_content,
                title=self.unset_to_none(attributes.title),
                rendering_layouts=rendering_layouts,
                outline_numbering=self.unset_to_none(
                    attributes.uses_outline_numbering
                ),
                outline_numbering_prefix=(
                    self.unset_to_none(attributes.outline_numbering.prefix)
                    if attributes.outline_numbering
                    else None
                ),
                additional_properties=attributes.additional_properties or {},
                structure_link_role=self.unset_to_none(
                    attributes.structure_link_role
                ),
            )
        return None

    def _pre_batching_grouping(
        self, items: list[dm.Document]
    ) -> t.Generator[list[dm.Document], None, None]:
        for _, group in itertools.groupby(items, lambda x: x.module_folder):
            yield list(group)

    def _update(self, to_update: list[dm.Document]) -> None:
        assert len(to_update) == 1, "Expected only one item"
        assert to_update[0].module_folder
        assert to_update[0].module_name
        res = patch_document.sync_detailed(
            project_id=self._project_id,
            space_id=to_update[0].module_folder,
            document_name=to_update[0].module_name,
            client=self._client.client,
            body=self._prepare_patch_request(to_update[0]),
        )

        self._raise_on_error(res)

    async def _async_update(self, to_update: list[dm.Document]) -> None:
        assert len(to_update) == 1, "Expected only one item"
        assert to_update[0].module_folder
        assert to_update[0].module_name
        res = await patch_document.asyncio_detailed(
            project_id=self._project_id,
            space_id=to_update[0].module_folder,
            document_name=to_update[0].module_name,
            client=self._client.client,
            body=self._prepare_patch_request(to_update[0]),
        )

        self._raise_on_error(res)

    def _prepare_patch_request(
        self, to_update: dm.Document
    ) -> api_models.DocumentsSinglePatchRequest:
        assert to_update.module_folder is not None, "module folder must be set"
        assert to_update.module_name is not None, "module name must be set"

        if to_update.structure_link_role:
            logger.warning(
                "Changing the documents structure link role is not supported."
            )

        attrs = api_models.DocumentsSinglePatchRequestDataAttributes(
            home_page_content=(
                api_models.DocumentsSinglePatchRequestDataAttributesHomePageContent(
                    type_=api_models.DocumentsSinglePatchRequestDataAttributesHomePageContentType(
                        to_update.home_page_content.type
                    ),
                    value=to_update.home_page_content.value or "",
                )
                if to_update.home_page_content
                else oa_types.UNSET
            ),
            status=to_update.status or oa_types.UNSET,
            title=to_update.title or oa_types.UNSET,
            type_=to_update.type or oa_types.UNSET,
            rendering_layouts=(
                [
                    api_models.DocumentsSinglePatchRequestDataAttributesRenderingLayoutsItem(
                        label=layout.label or oa_types.UNSET,
                        layouter=(
                            layout.layouter.value
                            if layout.layouter is not None
                            else oa_types.UNSET
                        ),
                        type_=layout.type or oa_types.UNSET,
                        properties=(
                            [
                                api_models.DocumentsSinglePatchRequestDataAttributesRenderingLayoutsItemPropertiesItem.from_dict(
                                    p
                                )
                                for p in layout.properties.serialize()
                            ]
                            if layout.properties
                            else oa_types.UNSET
                        ),
                    )
                    for layout in to_update.rendering_layouts
                ]
                if to_update.rendering_layouts
                else oa_types.UNSET
            ),
            uses_outline_numbering=to_update.outline_numbering
            or oa_types.UNSET,
            outline_numbering=(
                api_models.DocumentsSinglePatchRequestDataAttributesOutlineNumbering(
                    prefix=to_update.outline_numbering_prefix
                )
                if to_update.outline_numbering_prefix
                else oa_types.UNSET
            ),
        )

        attrs.additional_properties.update(
            to_update.additional_properties or {}
        )

        return api_models.DocumentsSinglePatchRequest(
            data=api_models.DocumentsSinglePatchRequestData(
                api_models.DocumentsSinglePatchRequestDataType.DOCUMENTS,
                id=f"{self._project_id}/{to_update.module_folder}/{to_update.module_name}",
                attributes=attrs,
            )
        )

    def _create(self, items: list[dm.Document]) -> None:
        assert items[0].module_folder
        res = post_documents.sync_detailed(
            self._project_id,
            items[0].module_folder,
            client=self._client.client,
            body=self._prepare_document_post_request(items),
        )

        self._raise_on_error(res)

    async def _async_create(self, items: list[dm.Document]) -> None:
        assert items[0].module_folder
        res = await post_documents.asyncio_detailed(
            self._project_id,
            items[0].module_folder,
            client=self._client.client,
            body=self._prepare_document_post_request(items),
        )

        self._raise_on_error(res)

    def _prepare_document_post_request(
        self, items: list[dm.Document]
    ) -> api_models.DocumentsListPostRequest:
        # due to grouping in _split_into_batches all module folders are equal
        assert items[0].module_folder is not None, "module folder must be set"
        return api_models.DocumentsListPostRequest(
            # pylint: disable=line-too-long
            data=[
                api_models.DocumentsListPostRequestDataItem(
                    type_=api_models.DocumentsListPostRequestDataItemType.DOCUMENTS,
                    attributes=self._build_create_attributes(document),
                )
                for document in items
            ]
        )

    def _build_create_attributes(
        self, document: dm.Document
    ) -> api_models.DocumentsListPostRequestDataItemAttributes:
        attrs = api_models.DocumentsListPostRequestDataItemAttributes(
            home_page_content=(
                api_models.DocumentsListPostRequestDataItemAttributesHomePageContent(
                    type_=api_models.DocumentsListPostRequestDataItemAttributesHomePageContentType(
                        document.home_page_content.type
                    ),
                    value=document.home_page_content.value or "",
                )
                if document.home_page_content
                else oa_types.UNSET
            ),
            module_name=document.module_name or oa_types.UNSET,
            status=document.status or oa_types.UNSET,
            title=document.title or oa_types.UNSET,
            type_=document.type or oa_types.UNSET,
            rendering_layouts=(
                [
                    api_models.DocumentsListPostRequestDataItemAttributesRenderingLayoutsItem(
                        label=layout.label or oa_types.UNSET,
                        layouter=(
                            layout.layouter.value
                            if layout.layouter is not None
                            else oa_types.UNSET
                        ),
                        type_=layout.type or oa_types.UNSET,
                        properties=(
                            [
                                api_models.DocumentsListPostRequestDataItemAttributesRenderingLayoutsItemPropertiesItem.from_dict(
                                    p
                                )
                                for p in layout.properties.serialize()
                            ]
                            if layout.properties
                            else oa_types.UNSET
                        ),
                    )
                    for layout in document.rendering_layouts
                ]
                if document.rendering_layouts
                else oa_types.UNSET
            ),
            uses_outline_numbering=document.outline_numbering
            or oa_types.UNSET,
            outline_numbering=(
                api_models.DocumentsListPostRequestDataItemAttributesOutlineNumbering(
                    prefix=document.outline_numbering_prefix
                )
                if document.outline_numbering_prefix
                else oa_types.UNSET
            ),
            structure_link_role=document.structure_link_role or oa_types.UNSET,
        )
        attrs.additional_properties.update(
            document.additional_properties or {}
        )
        return attrs



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
from collections.abc import Mapping
from io import BytesIO
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from .. import types
from ..types import UNSET, File, Unset

if TYPE_CHECKING:
    from ..models.teststepresult_attachments_single_patch_request import (
        TeststepresultAttachmentsSinglePatchRequest,
    )


T = TypeVar("T", bound="PatchTestStepResultAttachmentsRequestBody")


@_attrs_define
class PatchTestStepResultAttachmentsRequestBody:
    """
    Attributes:
        resource (TeststepresultAttachmentsSinglePatchRequest):
        content (Union[Unset, File]): attachments content
    """

    resource: "TeststepresultAttachmentsSinglePatchRequest"
    content: Union[Unset, File] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        resource = self.resource.to_dict()

        content: Union[Unset, types.FileTypes] = UNSET
        if not isinstance(self.content, Unset):
            content = self.content.to_tuple()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "resource": resource,
            }
        )
        if content is not UNSET:
            field_dict["content"] = content

        return field_dict

    def to_multipart(self) -> types.RequestFiles:
        files: types.RequestFiles = []

        files.append(
            (
                "resource",
                (
                    None,
                    json.dumps(self.resource.to_dict()).encode(),
                    "text/plain",
                ),
            )
        )

        if not isinstance(self.content, Unset):
            files.append(("content", self.content.to_tuple()))

        for prop_name, prop in self.additional_properties.items():
            files.append((prop_name, (None, str(prop).encode(), "text/plain")))

        return files

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.teststepresult_attachments_single_patch_request import (
            TeststepresultAttachmentsSinglePatchRequest,
        )

        d = dict(src_dict)
        resource = TeststepresultAttachmentsSinglePatchRequest.from_dict(
            d.pop("resource")
        )

        _content = d.pop("content", UNSET)
        content: Union[Unset, File]
        if isinstance(_content, Unset):
            content = UNSET
        else:
            content = File(payload=BytesIO(_content))

        patch_test_step_result_attachments_request_body_obj = cls(
            resource=resource,
            content=content,
        )

        patch_test_step_result_attachments_request_body_obj.additional_properties = d
        return patch_test_step_result_attachments_request_body_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.users_list_post_request_data_item_attributes_description import (
        UsersListPostRequestDataItemAttributesDescription,
    )


T = TypeVar("T", bound="UsersListPostRequestDataItemAttributes")


@_attrs_define
class UsersListPostRequestDataItemAttributes:
    """
    Attributes:
        id (str):  Example: MyUserId.
        description (Union[Unset, UsersListPostRequestDataItemAttributesDescription]):
        disabled_notifications (Union[Unset, bool]):
        email (Union[Unset, str]):  Example: Email.
        initials (Union[Unset, str]):  Example: Initials.
        name (Union[Unset, str]):  Example: Name.
    """

    id: str
    description: Union[
        Unset, "UsersListPostRequestDataItemAttributesDescription"
    ] = UNSET
    disabled_notifications: Union[Unset, bool] = UNSET
    email: Union[Unset, str] = UNSET
    initials: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        description: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.description, Unset):
            description = self.description.to_dict()

        disabled_notifications = self.disabled_notifications

        email = self.email

        initials = self.initials

        name = self.name

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if disabled_notifications is not UNSET:
            field_dict["disabledNotifications"] = disabled_notifications
        if email is not UNSET:
            field_dict["email"] = email
        if initials is not UNSET:
            field_dict["initials"] = initials
        if name is not UNSET:
            field_dict["name"] = name

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.users_list_post_request_data_item_attributes_description import (
            UsersListPostRequestDataItemAttributesDescription,
        )

        d = dict(src_dict)
        id = d.pop("id")

        _description = d.pop("description", UNSET)
        description: Union[
            Unset, UsersListPostRequestDataItemAttributesDescription
        ]
        if isinstance(_description, Unset):
            description = UNSET
        else:
            description = (
                UsersListPostRequestDataItemAttributesDescription.from_dict(
                    _description
                )
            )

        disabled_notifications = d.pop("disabledNotifications", UNSET)

        email = d.pop("email", UNSET)

        initials = d.pop("initials", UNSET)

        name = d.pop("name", UNSET)

        users_list_post_request_data_item_attributes_obj = cls(
            id=id,
            description=description,
            disabled_notifications=disabled_notifications,
            email=email,
            initials=initials,
            name=name,
        )

        users_list_post_request_data_item_attributes_obj.additional_properties = d
        return users_list_post_request_data_item_attributes_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.externallylinkedworkitems_single_get_response_data_meta_errors_item import (
        ExternallylinkedworkitemsSingleGetResponseDataMetaErrorsItem,
    )


T = TypeVar("T", bound="ExternallylinkedworkitemsSingleGetResponseDataMeta")


@_attrs_define
class ExternallylinkedworkitemsSingleGetResponseDataMeta:
    """
    Attributes:
        errors (Union[Unset, list['ExternallylinkedworkitemsSingleGetResponseDataMetaErrorsItem']]):
    """

    errors: Union[
        Unset,
        list["ExternallylinkedworkitemsSingleGetResponseDataMetaErrorsItem"],
    ] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        errors: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.errors, Unset):
            errors = []
            for errors_item_data in self.errors:
                errors_item = errors_item_data.to_dict()
                errors.append(errors_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if errors is not UNSET:
            field_dict["errors"] = errors

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.externallylinkedworkitems_single_get_response_data_meta_errors_item import (
            ExternallylinkedworkitemsSingleGetResponseDataMetaErrorsItem,
        )

        d = dict(src_dict)
        errors = []
        _errors = d.pop("errors", UNSET)
        for errors_item_data in _errors or []:
            errors_item = ExternallylinkedworkitemsSingleGetResponseDataMetaErrorsItem.from_dict(
                errors_item_data
            )

            errors.append(errors_item)

        externallylinkedworkitems_single_get_response_data_meta_obj = cls(
            errors=errors,
        )

        externallylinkedworkitems_single_get_response_data_meta_obj.additional_properties = d
        return externallylinkedworkitems_single_get_response_data_meta_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="TestrunCommentsSingleGetResponseLinks")


@_attrs_define
class TestrunCommentsSingleGetResponseLinks:
    """
    Attributes:
        self_ (Union[Unset, str]):  Example: server-host-name/application-
            path/projects/MyProjectId/testruns/MyTestRunId/comments/MyCommentId?revision=1234.
    """

    self_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        self_ = self.self_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if self_ is not UNSET:
            field_dict["self"] = self_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        self_ = d.pop("self", UNSET)

        testrun_comments_single_get_response_links_obj = cls(
            self_=self_,
        )

        testrun_comments_single_get_response_links_obj.additional_properties = d
        return testrun_comments_single_get_response_links_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from http import HTTPStatus
from typing import Any, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.errors import Errors
from ...models.linkedworkitems_single_patch_request import (
    LinkedworkitemsSinglePatchRequest,
)
from ...types import Response


def _get_kwargs(
    project_id: str,
    work_item_id: str,
    role_id: str,
    target_project_id: str,
    linked_work_item_id: str,
    *,
    body: LinkedworkitemsSinglePatchRequest,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "patch",
        "url": f"/projects/{project_id}/workitems/{work_item_id}/linkedworkitems/{role_id}/{target_project_id}/{linked_work_item_id}",
    }

    _kwargs["json"] = body.to_dict()

    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Union[Any, Errors] | None:
    if response.status_code == 204:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == 400:
        response_400 = Errors.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = Errors.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = Errors.from_dict(response.json())

        return response_403
    if response.status_code == 404:
        response_404 = Errors.from_dict(response.json())

        return response_404
    if response.status_code == 409:
        response_409 = Errors.from_dict(response.json())

        return response_409
    if response.status_code == 413:
        response_413 = Errors.from_dict(response.json())

        return response_413
    if response.status_code == 415:
        response_415 = Errors.from_dict(response.json())

        return response_415
    if response.status_code == 500:
        response_500 = Errors.from_dict(response.json())

        return response_500
    if response.status_code == 503:
        response_503 = Errors.from_dict(response.json())

        return response_503
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, Errors]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    project_id: str,
    work_item_id: str,
    role_id: str,
    target_project_id: str,
    linked_work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsSinglePatchRequest,
) -> Response[Union[Any, Errors]]:
    """Updates the specified Linked Work Item.

     Updates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        role_id (str):
        target_project_id (str):
        linked_work_item_id (str):
        body (LinkedworkitemsSinglePatchRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, Errors]]
    """

    kwargs = _get_kwargs(
        project_id=project_id,
        work_item_id=work_item_id,
        role_id=role_id,
        target_project_id=target_project_id,
        linked_work_item_id=linked_work_item_id,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    project_id: str,
    work_item_id: str,
    role_id: str,
    target_project_id: str,
    linked_work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsSinglePatchRequest,
) -> Union[Any, Errors] | None:
    """Updates the specified Linked Work Item.

     Updates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        role_id (str):
        target_project_id (str):
        linked_work_item_id (str):
        body (LinkedworkitemsSinglePatchRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, Errors]
    """

    return sync_detailed(
        project_id=project_id,
        work_item_id=work_item_id,
        role_id=role_id,
        target_project_id=target_project_id,
        linked_work_item_id=linked_work_item_id,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    project_id: str,
    work_item_id: str,
    role_id: str,
    target_project_id: str,
    linked_work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsSinglePatchRequest,
) -> Response[Union[Any, Errors]]:
    """Updates the specified Linked Work Item.

     Updates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        role_id (str):
        target_project_id (str):
        linked_work_item_id (str):
        body (LinkedworkitemsSinglePatchRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, Errors]]
    """

    kwargs = _get_kwargs(
        project_id=project_id,
        work_item_id=work_item_id,
        role_id=role_id,
        target_project_id=target_project_id,
        linked_work_item_id=linked_work_item_id,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    project_id: str,
    work_item_id: str,
    role_id: str,
    target_project_id: str,
    linked_work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsSinglePatchRequest,
) -> Union[Any, Errors] | None:
    """Updates the specified Linked Work Item.

     Updates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        role_id (str):
        target_project_id (str):
        linked_work_item_id (str):
        body (LinkedworkitemsSinglePatchRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, Errors]
    """

    return (
        await asyncio_detailed(
            project_id=project_id,
            work_item_id=work_item_id,
            role_id=role_id,
            target_project_id=target_project_id,
            linked_work_item_id=linked_work_item_id,
            client=client,
            body=body,
        )
    ).parsed



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.enumerations_list_post_request_data_item_type import (
    EnumerationsListPostRequestDataItemType,
)
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.enumerations_list_post_request_data_item_attributes import (
        EnumerationsListPostRequestDataItemAttributes,
    )


T = TypeVar("T", bound="EnumerationsListPostRequestDataItem")


@_attrs_define
class EnumerationsListPostRequestDataItem:
    """
    Attributes:
        type_ (Union[Unset, EnumerationsListPostRequestDataItemType]):
        attributes (Union[Unset, EnumerationsListPostRequestDataItemAttributes]):
    """

    type_: Union[Unset, EnumerationsListPostRequestDataItemType] = UNSET
    attributes: Union[
        Unset, "EnumerationsListPostRequestDataItemAttributes"
    ] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        type_: Union[Unset, str] = UNSET
        if not isinstance(self.type_, Unset):
            type_ = self.type_.value

        attributes: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if type_ is not UNSET:
            field_dict["type"] = type_
        if attributes is not UNSET:
            field_dict["attributes"] = attributes

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.enumerations_list_post_request_data_item_attributes import (
            EnumerationsListPostRequestDataItemAttributes,
        )

        d = dict(src_dict)
        _type_ = d.pop("type", UNSET)
        type_: Union[Unset, EnumerationsListPostRequestDataItemType]
        if isinstance(_type_, Unset):
            type_ = UNSET
        else:
            type_ = EnumerationsListPostRequestDataItemType(_type_)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, EnumerationsListPostRequestDataItemAttributes]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = (
                EnumerationsListPostRequestDataItemAttributes.from_dict(
                    _attributes
                )
            )

        enumerations_list_post_request_data_item_obj = cls(
            type_=type_,
            attributes=attributes,
        )

        enumerations_list_post_request_data_item_obj.additional_properties = d
        return enumerations_list_post_request_data_item_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.collections_single_get_response_data_relationships_author import (
        CollectionsSingleGetResponseDataRelationshipsAuthor,
    )
    from ..models.collections_single_get_response_data_relationships_documents import (
        CollectionsSingleGetResponseDataRelationshipsDocuments,
    )
    from ..models.collections_single_get_response_data_relationships_downstream_collections import (
        CollectionsSingleGetResponseDataRelationshipsDownstreamCollections,
    )
    from ..models.collections_single_get_response_data_relationships_project import (
        CollectionsSingleGetResponseDataRelationshipsProject,
    )
    from ..models.collections_single_get_response_data_relationships_reused_from import (
        CollectionsSingleGetResponseDataRelationshipsReusedFrom,
    )
    from ..models.collections_single_get_response_data_relationships_rich_pages import (
        CollectionsSingleGetResponseDataRelationshipsRichPages,
    )
    from ..models.collections_single_get_response_data_relationships_upstream_collections import (
        CollectionsSingleGetResponseDataRelationshipsUpstreamCollections,
    )


T = TypeVar("T", bound="CollectionsSingleGetResponseDataRelationships")


@_attrs_define
class CollectionsSingleGetResponseDataRelationships:
    """
    Attributes:
        author (Union[Unset, CollectionsSingleGetResponseDataRelationshipsAuthor]):
        documents (Union[Unset, CollectionsSingleGetResponseDataRelationshipsDocuments]):
        downstream_collections (Union[Unset, CollectionsSingleGetResponseDataRelationshipsDownstreamCollections]):
        project (Union[Unset, CollectionsSingleGetResponseDataRelationshipsProject]):
        reused_from (Union[Unset, CollectionsSingleGetResponseDataRelationshipsReusedFrom]):
        rich_pages (Union[Unset, CollectionsSingleGetResponseDataRelationshipsRichPages]):
        upstream_collections (Union[Unset, CollectionsSingleGetResponseDataRelationshipsUpstreamCollections]):
    """

    author: Union[
        Unset, "CollectionsSingleGetResponseDataRelationshipsAuthor"
    ] = UNSET
    documents: Union[
        Unset, "CollectionsSingleGetResponseDataRelationshipsDocuments"
    ] = UNSET
    downstream_collections: Union[
        Unset,
        "CollectionsSingleGetResponseDataRelationshipsDownstreamCollections",
    ] = UNSET
    project: Union[
        Unset, "CollectionsSingleGetResponseDataRelationshipsProject"
    ] = UNSET
    reused_from: Union[
        Unset, "CollectionsSingleGetResponseDataRelationshipsReusedFrom"
    ] = UNSET
    rich_pages: Union[
        Unset, "CollectionsSingleGetResponseDataRelationshipsRichPages"
    ] = UNSET
    upstream_collections: Union[
        Unset,
        "CollectionsSingleGetResponseDataRelationshipsUpstreamCollections",
    ] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> dict[str, Any]:
        author: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.author, Unset):
            author = self.author.to_dict()

        documents: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.documents, Unset):
            documents = self.documents.to_dict()

        downstream_collections: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.downstream_collections, Unset):
            downstream_collections = self.downstream_collections.to_dict()

        project: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.project, Unset):
            project = self.project.to_dict()

        reused_from: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.reused_from, Unset):
            reused_from = self.reused_from.to_dict()

        rich_pages: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.rich_pages, Unset):
            rich_pages = self.rich_pages.to_dict()

        upstream_collections: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.upstream_collections, Unset):
            upstream_collections = self.upstream_collections.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if author is not UNSET:
            field_dict["author"] = author
        if documents is not UNSET:
            field_dict["documents"] = documents
        if downstream_collections is not UNSET:
            field_dict["downstreamCollections"] = downstream_collections
        if project is not UNSET:
            field_dict["project"] = project
        if reused_from is not UNSET:
            field_dict["reusedFrom"] = reused_from
        if rich_pages is not UNSET:
            field_dict["richPages"] = rich_pages
        if upstream_collections is not UNSET:
            field_dict["upstreamCollections"] = upstream_collections

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.collections_single_get_response_data_relationships_author import (
            CollectionsSingleGetResponseDataRelationshipsAuthor,
        )
        from ..models.collections_single_get_response_data_relationships_documents import (
            CollectionsSingleGetResponseDataRelationshipsDocuments,
        )
        from ..models.collections_single_get_response_data_relationships_downstream_collections import (
            CollectionsSingleGetResponseDataRelationshipsDownstreamCollections,
        )
        from ..models.collections_single_get_response_data_relationships_project import (
            CollectionsSingleGetResponseDataRelationshipsProject,
        )
        from ..models.collections_single_get_response_data_relationships_reused_from import (
            CollectionsSingleGetResponseDataRelationshipsReusedFrom,
        )
        from ..models.collections_single_get_response_data_relationships_rich_pages import (
            CollectionsSingleGetResponseDataRelationshipsRichPages,
        )
        from ..models.collections_single_get_response_data_relationships_upstream_collections import (
            CollectionsSingleGetResponseDataRelationshipsUpstreamCollections,
        )

        d = dict(src_dict)
        _author = d.pop("author", UNSET)
        author: Union[
            Unset, CollectionsSingleGetResponseDataRelationshipsAuthor
        ]
        if isinstance(_author, Unset):
            author = UNSET
        else:
            author = (
                CollectionsSingleGetResponseDataRelationshipsAuthor.from_dict(
                    _author
                )
            )

        _documents = d.pop("documents", UNSET)
        documents: Union[
            Unset, CollectionsSingleGetResponseDataRelationshipsDocuments
        ]
        if isinstance(_documents, Unset):
            documents = UNSET
        else:
            documents = CollectionsSingleGetResponseDataRelationshipsDocuments.from_dict(
                _documents
            )

        _downstream_collections = d.pop("downstreamCollections", UNSET)
        downstream_collections: Union[
            Unset,
            CollectionsSingleGetResponseDataRelationshipsDownstreamCollections,
        ]
        if isinstance(_downstream_collections, Unset):
            downstream_collections = UNSET
        else:
            downstream_collections = CollectionsSingleGetResponseDataRelationshipsDownstreamCollections.from_dict(
                _downstream_collections
            )

        _project = d.pop("project", UNSET)
        project: Union[
            Unset, CollectionsSingleGetResponseDataRelationshipsProject
        ]
        if isinstance(_project, Unset):
            project = UNSET
        else:
            project = (
                CollectionsSingleGetResponseDataRelationshipsProject.from_dict(
                    _project
                )
            )

        _reused_from = d.pop("reusedFrom", UNSET)
        reused_from: Union[
            Unset, CollectionsSingleGetResponseDataRelationshipsReusedFrom
        ]
        if isinstance(_reused_from, Unset):
            reused_from = UNSET
        else:
            reused_from = CollectionsSingleGetResponseDataRelationshipsReusedFrom.from_dict(
                _reused_from
            )

        _rich_pages = d.pop("richPages", UNSET)
        rich_pages: Union[
            Unset, CollectionsSingleGetResponseDataRelationshipsRichPages
        ]
        if isinstance(_rich_pages, Unset):
            rich_pages = UNSET
        else:
            rich_pages = CollectionsSingleGetResponseDataRelationshipsRichPages.from_dict(
                _rich_pages
            )

        _upstream_collections = d.pop("upstreamCollections", UNSET)
        upstream_collections: Union[
            Unset,
            CollectionsSingleGetResponseDataRelationshipsUpstreamCollections,
        ]
        if isinstance(_upstream_collections, Unset):
            upstream_collections = UNSET
        else:
            upstream_collections = CollectionsSingleGetResponseDataRelationshipsUpstreamCollections.from_dict(
                _upstream_collections
            )

        collections_single_get_response_data_relationships_obj = cls(
            author=author,
            documents=documents,
            downstream_collections=downstream_collections,
            project=project,
            reused_from=reused_from,
            rich_pages=rich_pages,
            upstream_collections=upstream_collections,
        )

        collections_single_get_response_data_relationships_obj.additional_properties = d
        return collections_single_get_response_data_relationships_obj

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json

import pytest_httpx

import polarion_rest_api_client as polarion_api
from tests.conftest import (
    TEST_TRUN_PARAM_CREATED_REQ_1,
    TEST_TRUN_PARAM_CREATED_REQ_2,
    TEST_TRUN_PARAM_CREATED_RES,
    TEST_TRUN_PARAM_DELETE_REQ_1,
    TEST_TRUN_PARAM_DELETE_REQ_2,
    TEST_TRUN_PARAM_NEXT_RES,
    TEST_TRUN_PARAM_NO_NEXT_RES,
)


def test_get_test_params_trun_multi_page(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_TRUN_PARAM_NEXT_RES, encoding="utf8") as f:
        content = json.load(f)
        httpx_mock.add_response(json=content)
        httpx_mock.add_response(json=content)

    with open(TEST_TRUN_PARAM_NO_NEXT_RES, encoding="utf8") as f:
        httpx_mock.add_response(json=json.load(f))

    params: list[polarion_api.TestRunParameter] = (
        client.test_runs.parameters.get_all(
            "123", fields={"teststeps": "@basic"}
        )
    )

    query = {
        "page[size]": "100",
        "page[number]": "1",
        "fields[teststeps]": "@basic",
    }
    reqs = httpx_mock.get_requests()

    assert len(reqs) == 3
    assert reqs[0].method == "GET"
    assert dict(reqs[0].url.params) == query
    query["page[number]"] = "2"
    assert dict(reqs[1].url.params) == query
    query["page[number]"] = "3"
    assert dict(reqs[2].url.params) == query
    assert all(
        req.url.path == "/api/projects/PROJ/testruns/123/testparameters"
        for req in reqs
    )
    assert len(params) == 6
    assert params.pop(5).name == "Another Name"
    assert all(
        param.test_run_id == "123"
        and param.name == "Example Test Parameter value"
        and param.value == "Example Test Parameter value"
        for param in params
    )


def test_get_test_params_trecord_multi_page(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_TRUN_PARAM_NEXT_RES, encoding="utf8") as f:
        content = json.load(f)
        httpx_mock.add_response(json=content)

    with open(TEST_TRUN_PARAM_NO_NEXT_RES, encoding="utf8") as f:
        httpx_mock.add_response(json=json.load(f))

    params: list[polarion_api.TestRecordParameter] = (
        client.test_runs.records.parameters.get_all(
            polarion_api.TestRecord("TRUN", "PROJ", "TCID", iteration=0),
            fields={"teststeps": "@basic"},
        )
    )

    query = {
        "page[size]": "100",
        "page[number]": "1",
        "fields[teststeps]": "@basic",
    }
    reqs = httpx_mock.get_requests()

    assert len(reqs) == 2
    assert reqs[0].method == "GET"
    assert (
        reqs[0].url.path
        == "/api/projects/PROJ/testruns/TRUN/testrecords/PROJ/TCID/0/testparameters"
    )
    assert dict(reqs[0].url.params) == query
    query["page[number]"] = "2"
    assert reqs[1].method == "GET"
    assert (
        reqs[1].url.path
        == "/api/projects/PROJ/testruns/TRUN/testrecords/PROJ/TCID/0/testparameters"
    )
    assert dict(reqs[1].url.params) == query
    assert len(params) == 4
    assert params.pop(3).name == "Another Name"
    assert all(
        param.test_record
        == polarion_api.TestRecord("TRUN", "PROJ", "TCID", iteration=0)
        and param.name == "Example Test Parameter value"
        and param.value == "Example Test Parameter value"
        for param in params
    )


def test_create_parameters_trun(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_TRUN_PARAM_CREATED_RES, encoding="utf8") as f:
        content = json.load(f)
        httpx_mock.add_response(201, json=content)
        httpx_mock.add_response(201, json=content)

    with open(TEST_TRUN_PARAM_CREATED_REQ_1, encoding="utf8") as f:
        expected_req_1 = json.load(f)

    with open(TEST_TRUN_PARAM_CREATED_REQ_2, encoding="utf8") as f:
        expected_req_2 = json.load(f)

    test_run_id = "asdfg"
    test_run_id_1 = "qwertz"

    tp_1 = polarion_api.TestRunParameter(test_run_id, "name1", "value1")
    tp_2 = polarion_api.TestRunParameter(test_run_id_1, "name2", "value2")
    tp_3 = polarion_api.TestRunParameter(test_run_id, "name3", "value3")

    client.test_runs.parameters.create([tp_1, tp_2, tp_3])

    reqs = httpx_mock.get_requests()
    req_data_1 = json.loads(reqs[0].content.decode("utf-8"))
    req_data_2 = json.loads(reqs[1].content.decode("utf-8"))
    assert len(reqs) == 2
    assert (
        reqs[0].url.path == "/api/projects/PROJ/testruns/asdfg/testparameters"
    )
    assert (
        reqs[1].url.path == "/api/projects/PROJ/testruns/qwertz/testparameters"
    )
    assert reqs[0].method == "POST"
    assert reqs[1].method == "POST"
    assert req_data_1 == expected_req_1
    assert req_data_2 == expected_req_2


def test_create_parameters_trecord(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_TRUN_PARAM_CREATED_RES, encoding="utf8") as f:
        content = json.load(f)
        httpx_mock.add_response(201, json=content)
        httpx_mock.add_response(201, json=content)

    with open(TEST_TRUN_PARAM_CREATED_REQ_1, encoding="utf8") as f:
        expected_req_1 = json.load(f)

    with open(TEST_TRUN_PARAM_CREATED_REQ_2, encoding="utf8") as f:
        expected_req_2 = json.load(f)

    test_record_1 = polarion_api.TestRecord("TRUN", "PROJ", "TC1", iteration=0)
    test_record_2 = polarion_api.TestRecord("TRUN", "PROJ", "TC1", iteration=1)

    tp_1 = polarion_api.TestRecordParameter(test_record_1, "name1", "value1")
    tp_2 = polarion_api.TestRecordParameter(test_record_2, "name2", "value2")
    tp_3 = polarion_api.TestRecordParameter(test_record_1, "name3", "value3")

    client.test_runs.records.parameters.create([tp_1, tp_2, tp_3])

    reqs = httpx_mock.get_requests()
    req_data_1 = json.loads(reqs[0].content.decode("utf-8"))
    req_data_2 = json.loads(reqs[1].content.decode("utf-8"))
    assert len(reqs) == 2
    assert (
        reqs[0].url.path
        == "/api/projects/PROJ/testruns/TRUN/testrecords/PROJ/TC1/0/testparameters"
    )
    assert (
        reqs[1].url.path
        == "/api/projects/PROJ/testruns/TRUN/testrecords/PROJ/TC1/1/testparameters"
    )
    assert reqs[0].method == "POST"
    assert reqs[1].method == "POST"
    assert req_data_1 == expected_req_1
    assert req_data_2 == expected_req_2


def test_delete_parameters_trun(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    httpx_mock.add_response(204)
    httpx_mock.add_response(204)

    with open(TEST_TRUN_PARAM_DELETE_REQ_1, encoding="utf8") as f:
        expected_req_1 = json.load(f)

    with open(TEST_TRUN_PARAM_DELETE_REQ_2, encoding="utf8") as f:
        expected_req_2 = json.load(f)

    tp_1 = polarion_api.TestRunParameter("TRUN", "name1", "value1")
    tp_2 = polarion_api.TestRunParameter("TRUN2", "name1", "value2")
    tp_3 = polarion_api.TestRunParameter("TRUN", "name2", "value1")

    client.test_runs.parameters.delete([tp_1, tp_2, tp_3])

    reqs = httpx_mock.get_requests()
    req_data_1 = json.loads(reqs[0].content.decode("utf-8"))
    req_data_2 = json.loads(reqs[1].content.decode("utf-8"))
    assert len(reqs) == 2
    assert (
        reqs[0].url.path == "/api/projects/PROJ/testruns/TRUN/testparameters"
    )
    assert (
        reqs[1].url.path == "/api/projects/PROJ/testruns/TRUN2/testparameters"
    )
    assert reqs[0].method == "DELETE"
    assert reqs[1].method == "DELETE"
    assert req_data_1 == expected_req_1
    assert req_data_2 == expected_req_2


def test_delete_parameters_trecord(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    httpx_mock.add_response(204)
    httpx_mock.add_response(204)
    httpx_mock.add_response(204)

    test_record_1 = polarion_api.TestRecord("TRUN", "PROJ", "TC1", iteration=0)
    test_record_2 = polarion_api.TestRecord("TRUN", "PROJ", "TC1", iteration=1)

    tp_1 = polarion_api.TestRecordParameter(test_record_1, "name1", "value1")
    tp_2 = polarion_api.TestRecordParameter(test_record_2, "name2", "value2")
    tp_3 = polarion_api.TestRecordParameter(test_record_1, "name3", "value3")

    client.test_runs.records.parameters.delete([tp_1, tp_2, tp_3])

    reqs = httpx_mock.get_requests()
    assert len(reqs) == 3
    assert (
        reqs[0].url.path
        == "/api/projects/PROJ/testruns/TRUN/testrecords/PROJ/TC1/0/testparameters/name1"
    )
    assert (
        reqs[1].url.path
        == "/api/projects/PROJ/testruns/TRUN/testrecords/PROJ/TC1/0/testparameters/name3"
    )
    assert (
        reqs[2].url.path
        == "/api/projects/PROJ/testruns/TRUN/testrecords/PROJ/TC1/1/testparameters/name2"
    )
    assert reqs[0].method == "DELETE"
    assert reqs[1].method == "DELETE"
    assert reqs[2].method == "DELETE"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Base classes for client implementations on project Level."""

import abc
import asyncio
import datetime
import functools
import http
import inspect
import logging
import random
import time
import typing as t

import httpx

from polarion_rest_api_client import data_models as dm
from polarion_rest_api_client import errors
from polarion_rest_api_client.open_api_client import models as api_models
from polarion_rest_api_client.open_api_client import types as oa_types

R = t.TypeVar("R")

if t.TYPE_CHECKING:
    from polarion_rest_api_client import client as polarion_client

T = t.TypeVar("T")
ST = t.TypeVar("ST", bound=dm.StatusItem)
logger = logging.getLogger(__name__)
_max_retries: t.Final = 5

_NON_RETRIABLE_ERRORS = {
    http.HTTPStatus.BAD_REQUEST,
    http.HTTPStatus.IM_A_TEAPOT,
    http.HTTPStatus.NOT_FOUND,
}

UT = t.TypeVar("UT", str, int, float, datetime.datetime, bool, None)
NT = t.TypeVar("NT", str, int, float, datetime.datetime, bool, oa_types.Unset)


def _compute_backoff(sleeptime: float, attempt: int) -> float:
    return sleeptime * 3 + random.uniform(0.5, (attempt + 1) / 2)


class BaseClient(t.Generic[T]):
    """The overall base client for all project related clients."""

    _retry_methods: t.ClassVar[set[str]] = (
        set()
    )  # methods which can be retried

    def __init__(
        self, project_id: str, client: "polarion_client.PolarionClient"
    ):
        self._project_id = project_id
        self._client = client
        self._all_retry_methods: set[str] = set()
        for base in self.__class__.__mro__:
            if hasattr(base, "_retry_methods"):
                self._all_retry_methods.update(base._retry_methods)

    def __getattribute__(self, name: str) -> t.Any:
        """Retry method calls defined in _retry_methods."""
        attr = super().__getattribute__(name)
        retry_methods = super().__getattribute__("_all_retry_methods")
        if name in retry_methods and callable(attr):
            return functools.partial(
                super().__getattribute__("_retry_on_error"), attr
            )

        return attr

    def _handle_text_content(
        self,
        polarion_content: (
            api_models.DocumentsSingleGetResponseDataAttributesHomePageContent  # pylint: disable=line-too-long
            | api_models.TestrecordsListGetResponseDataItemAttributesComment
            | api_models.TestrunsListGetResponseDataItemAttributesHomePageContent
            | oa_types.Unset
        ),
    ) -> dm.TextContent | None:
        if not polarion_content:
            return None

        return dm.TextContent(
            type=(
                str(polarion_content.type_) if polarion_content.type_ else None
            ),
            value=polarion_content.value or None,
        )

    def _build_sparse_fields(
        self, fields_dict: dict[str, str]
    ) -> api_models.SparseFields | oa_types.Unset:
        """Build the SparseFields object based on a dict.

        Ensure that every key follow the pattern 'fields[XXX]'.
        """
        new_field_dict: dict[str, str] = {}
        for key, value in fields_dict.items():
            if key.startswith("fields["):
                new_field_dict[key] = value
            else:
                new_field_dict[f"fields[{key}]"] = value
        return api_models.SparseFields.from_dict(new_field_dict)

    @t.overload
    def unset_to_none(self, value: oa_types.Unset) -> None:
        """Return None if value is Unset, else the value."""

    @t.overload
    def unset_to_none(self, value: UT) -> UT:
        """Return None if value is Unset, else the value."""

    def unset_to_none(self, value: t.Any) -> t.Any:
        """Return None if value is Unset, else the value."""
        if isinstance(value, oa_types.Unset):
            return None
        return value

    @t.overload
    def none_to_unset(self, value: None) -> oa_types.Unset:
        """Return UNSET if value is None, else the value."""

    @t.overload
    def none_to_unset(self, value: NT) -> NT:
        """Return UNSET if value is None, else the value."""

    def none_to_unset(self, value: t.Any) -> t.Any:
        """Return UNSET if value is None, else the value."""
        if value is None:
            return oa_types.UNSET
        return value

    def _raise_on_error(self, response: oa_types.Response) -> None:
        """Raise a PolarionApiBaseException, if the response has errors."""

        def unexpected_error() -> errors.PolarionApiUnexpectedException:
            return errors.PolarionApiUnexpectedException(
                response.status_code, response.content
            )

        if response.status_code not in range(400, 600):
            return

        if (
            isinstance(response.parsed, api_models.Errors)
            and response.parsed.errors
        ):
            raise errors.PolarionApiException(
                response.status_code,
                *[
                    (
                        e.status,
                        e.detail,
                        (
                            e.source.pointer
                            if not (
                                isinstance(e.source, oa_types.Unset)
                                or e.source is None
                            )
                            else "No error pointer"
                        ),
                    )
                    for e in response.parsed.errors
                ],
            )
        raise unexpected_error()

    def _retry_on_error(
        self, call: t.Callable[..., R], *args: t.Any, **kwargs: t.Any
    ) -> R | t.Coroutine[t.Any, t.Any, R]:
        """Retry on errors with exponential backoff."""
        if inspect.iscoroutinefunction(call):
            return self._retry_async(call, *args, **kwargs)
        return self._retry_sync(call, *args, **kwargs)

    def _retry_sync(
        self, call: t.Callable[..., R], *args: t.Any, **kwargs: t.Any
    ) -> R:
        last_error: Exception | None = None
        sleeptime = 0.0
        for attempt in range(1, _max_retries + 1):
            try:
                return call(*args, **kwargs)
            except (errors.PolarionApiBaseException, httpx.HTTPError) as e:
                sleeptime = _compute_backoff(sleeptime, attempt)
                self._handle_tolerated_exception(e, sleeptime, attempt)
                last_error = e
                time.sleep(sleeptime)

        assert last_error is not None
        raise last_error

    def _retry_async(
        self,
        call: t.Callable[..., t.Coroutine[t.Any, t.Any, R]],
        *args: t.Any,
        **kwargs: t.Any,
    ) -> t.Coroutine[t.Any, t.Any, R]:
        async def wrapped() -> R:
            last_error: Exception | None = None
            sleeptime = 0.0
            for attempt in range(1, _max_retries + 1):
                try:
                    async with self._client.semaphore:
                        return await call(*args, **kwargs)
                except (errors.PolarionApiBaseException, httpx.HTTPError) as e:
                    sleeptime = _compute_backoff(sleeptime, attempt)
                    self._handle_tolerated_exception(e, sleeptime, attempt)
                    last_error = e
                    await asyncio.sleep(sleeptime)

            assert last_error is not None
            raise last_error

        return wrapped()

    def _handle_tolerated_exception(
        self,
        error: Exception,
        sleeptime: float,
        attempt: int,
    ) -> None:
        """Log the error without raising if it's a retryable error."""
        if (
            isinstance(error, errors.PolarionApiException)
            and error.args[0] in _NON_RETRIABLE_ERRORS
        ):
            raise error

        logger.warning(
            "Request failed, retrying after %.1fs (%d/%d): %s",
            sleeptime,
            attempt,
            _max_retries,
            error,
        )

    def _pre_batching_grouping(
        self, items: list[T]
    ) -> t.Generator[list[T], None, None]:
        """Group items before actually splitting them into batches.

        Needed if we need to first group by a specific value before we
        can split into batches.
        """
        yield items


class SingleGetClient(BaseClient[T], abc.ABC):
    """A client for items of a project, which can be created."""

    _retry_methods: t.ClassVar[set[str]] = {
        "get",
        "async_get",
    }

    @abc.abstractmethod
    def get(self, *args: t.Any, **kwargs: t.Any) -> T | None:
        """Get a specific single item."""

    @abc.abstractmethod
    async def async_get(self, *args: t.Any, **kwargs: t.Any) -> T | None:
        """Get a specific single item."""


class CreateClient(BaseClient[T], abc.ABC):
    """A client for items of a project, which can be created."""

    _retry_methods: t.ClassVar[set[str]] = {
        "_create",
        "_async_create",
    }
    _create_batch_size: int = 0

    @abc.abstractmethod
    def _create(self, items: list[T]) -> None: ...

    @abc.abstractmethod
    async def _async_create(self, items: list[T]) -> None: ...

    def _split_into_create_batches(
        self, items: list[T]
    ) -> t.Generator[list[T], None, None]:
        batch_size = self._create_batch_size or self._client.batch_size
        for it in self._pre_batching_grouping(items):
            for i in range(0, len(it), batch_size):
                yield it[i : i + batch_size]

    def create(self, items: T | list[T]) -> None:
        """Create one or multiple items."""
        if not isinstance(items, list):
            items = [items]

        for batch in self._split_into_create_batches(items):
            self._create(batch)

    async def async_create(self, items: T | list[T]) -> None:
        """Create one or multiple items."""
        if not isinstance(items, list):
            items = [items]

        for batch in self._split_into_create_batches(items):
            await self._async_create(batch)


class DeleteClient(BaseClient[T], abc.ABC):
    """A client for items of a project, which can be created."""

    _retry_methods: t.ClassVar[set[str]] = {
        "_delete",
        "_async_delete",
    }
    _delete_batch_size: int = 0

    def _split_into_delete_batches(
        self, items: list[T]
    ) -> t.Generator[list[T], None, None]:
        batch_size = self._delete_batch_size or self._client.batch_size
        for it in self._pre_batching_grouping(items):
            for i in range(0, len(it), batch_size):
                yield it[i : i + batch_size]

    @abc.abstractmethod
    def _delete(self, items: list[T]) -> None: ...

    @abc.abstractmethod
    async def _async_delete(self, items: list[T]) -> None: ...

    def delete(self, items: T | list[T]) -> None:
        """Delete one or multiple items."""
        if not isinstance(items, list):
            items = [items]
        for batch in self._split_into_delete_batches(items):
            self._delete(batch)

    async def async_delete(self, items: T | list[T]) -> None:
        """Delete one or multiple items."""
        if not isinstance(items, list):
            items = [items]
        for batch in self._split_into_delete_batches(items):
            await self._async_delete(batch)


class MultiGetClient(BaseClient[T], abc.ABC):
    """A client for items of a project, which can be created."""

    _retry_methods: t.ClassVar[set[str]] = {
        "get_multi",
        "async_get_multi",
    }

    @abc.abstractmethod
    def get_multi(
        self,
        *args: t.Any,
        page_size: int = 100,
        page_number: int = 1,
        **kwargs: t.Any,
    ) -> tuple[list[T], bool]:
        """Get multiple matching items for a specific page.

        In addition, a flag whether a next page is available is
        returned.
        """

    @abc.abstractmethod
    async def async_get_multi(
        self,
        *args: t.Any,
        page_size: int = 100,
        page_number: int = 1,
        **kwargs: t.Any,
    ) -> tuple[list[T], bool]:
        """Get multiple matching items for a specific page.

        In addition, a flag whether a next page is available is
        returned.
        """

    def get_all(self, *args: t.Any, **kwargs: t.Any) -> list[T]:
        """Return all matching items using get_multi with auto pagination."""
        page = 1
        items, next_page = self.get_multi(
            *args, page_size=self._client.page_size, page_number=page, **kwargs
        )
        while next_page:
            page += 1
            _items, next_page = self.get_multi(
                *args,
                page_size=self._client.page_size,
                page_number=page,
                **kwargs,
            )
            items += _items
        return items

    async def async_get_all(self, *args: t.Any, **kwargs: t.Any) -> list[T]:
        """Return all matching items using get_multi with auto pagination."""
        page = 1
        items, next_page = await self.async_get_multi(
            *args, page_size=self._client.page_size, page_number=page, **kwargs
        )
        while next_page:
            page += 1
            _items, next_page = await self.async_get_multi(
                *args,
                page_size=self._client.page_size,
                page_number=page,
                **kwargs,
            )
            items += _items
        return items


class UpdateClient(BaseClient[T], abc.ABC):
    """A client for items which can also be updated."""

    _retry_methods: t.ClassVar[set[str]] = {
        "_update",
        "_async_update",
    }
    _update_batch_size: int = 0

    def _split_into_update_batches(
        self, items: list[T]
    ) -> t.Generator[list[T], None, None]:
        batch_size = self._update_batch_size or self._client.batch_size
        for it in self._pre_batching_grouping(items):
            for i in range(0, len(it), batch_size):
                yield it[i : i + batch_size]

    @abc.abstractmethod
    def _update(self, to_update: list[T]) -> None: ...

    @abc.abstractmethod
    async def _async_update(self, to_update: list[T]) -> None: ...

    def update(self, items: T | list[T]) -> None:
        """Update the provided item or items."""
        if not isinstance(items, list):
            items = [items]

        for batch in self._split_into_update_batches(items):
            self._update(batch)

    async def async_update(self, items: T | list[T]) -> None:
        """Update the provided item or items."""
        if not isinstance(items, list):
            items = [items]

        for batch in self._split_into_update_batches(items):
            await self._async_update(batch)


class StatusItemClient(UpdateClient, DeleteClient, t.Generic[ST], abc.ABC):
    """A client for items, which have a status.

    We support to set a specific status for these instead of deleting
    them. This status has to be provided on initialization.
    """

    item_cls: type[ST]

    def __init__(
        self,
        project_id: str,
        client: "polarion_client.PolarionClient",
        delete_status: str | None = None,
    ):
        super().__init__(project_id, client)
        self.delete_status = delete_status

    def delete(self, items: ST | list[ST]) -> None:
        """Delete the item if no delete_status was set, else update status."""
        if self.delete_status is None:
            super().delete(items)
        else:
            delete_items = self._prepare_update_items(items)
            self.update(delete_items)

    async def async_delete(self, items: ST | list[ST]) -> None:
        """Delete the item if no delete_status was set, else update status."""
        if self.delete_status is None:
            await super().async_delete(items)
        else:
            delete_items = self._prepare_update_items(items)
            await self.async_update(delete_items)

    def _prepare_update_items(self, items: ST | list[ST]) -> list[ST]:
        if not isinstance(items, list):
            items = [items]
        delete_items: list[ST] = []
        for item in items:
            item.status = self.delete_status
            delete_items.append(
                self.item_cls(id=item.id, status=self.delete_status)
            )
        return delete_items



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Script to fix the specification and build code from it.

Usage: needs 2 args for execution. First one is either 'url' or 'path',
second one is the path to the Open API Spec. E.g.
./build_client_source.sh path /download/spec.json will take the spec
from the given path
"""

import json
import os
import pathlib
import re
import shutil
import subprocess
import sys
import tempfile

import httpx
import yaml

error_code_pattern = re.compile("[4,5][0-9]{2}")
script_path = pathlib.Path(os.path.dirname(os.path.realpath(__file__)))
rest_api_path = script_path.parent / "polarion_rest_api_client"
template_path = script_path / "custom_templates"
config_path = script_path / "config.yaml"

with open(config_path, encoding="utf-8") as config_file:
    config = yaml.safe_load(config_file)
    output_path = rest_api_path / config.get("package_name_override", "")


def get_and_fix_spec(src: str, path: str | os.PathLike):
    """Fix errors in the specification."""
    if src == "path":
        with open(path) as f:
            spec = json.load(f)
    elif src == "url":
        response = httpx.get(path)
        spec = response.json()
    else:
        raise Exception(
            "you have to provide a file or url keyword as 1st arg."
        )
    spec_paths = spec["paths"]
    if (
        octet_schema := spec_paths.get(
            "/projects/{projectId}/testruns/{testRunId}/actions/importXUnitTestResults",
            {},
        )
        .get("post", {})
        .get("requestBody", {})
        .get("content", {})
        .get("application/octet-stream", {})
        .get("schema")
    ) and octet_schema.get("type") == "object":
        octet_schema["type"] = "string"
        octet_schema["format"] = "binary"

    for spec_path in spec_paths.values():
        for operation_description in spec_path.values():
            if (
                responses := operation_description.get("responses")
            ) and "4XX-5XX" in responses:
                for code, resp in responses.items():
                    if error_code_pattern.fullmatch(code):
                        resp["content"] = responses["4XX-5XX"]["content"]
                del responses["4XX-5XX"]
    schemas = spec["components"]["schemas"]

    if (
        (
            downloads := schemas.get("jobsSingleGetResponse", {})
            .get("properties", {})
            .get("data", {})
            .get("properties", {})
            .get("links", {})
            .get("properties", {})
            .get("downloads")
        )
        and "items" not in downloads
        and downloads.get("type") == "array"
    ):
        downloads["items"] = {"type": "string"}

    if (
        (
            downloads := schemas.get("jobsSinglePostResponse", {})
            .get("properties", {})
            .get("data", {})
            .get("properties", {})
            .get("links", {})
            .get("properties", {})
            .get("downloads")
        )
        and "items" not in downloads
        and downloads.get("type") == "array"
    ):
        downloads["items"] = {"type": "string"}

    if (
        error_source := schemas.get("errors", {})
        .get("properties", {})
        .get("errors", {})
        .get("items", {})
        .get("properties", {})
        .get("source")
    ):
        error_source["nullable"] = True
        if resource := error_source.get("properties", {}).get("resource"):
            resource["nullable"] = True
    return spec


def generate_client(spec):
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        json.dump(spec, f)
        f.close()

        generator_path = shutil.which("openapi-python-client")
        assert generator_path is not None, (
            "Did not find openapi-python-client generator - "
            "please install dev requirements"
        )

        subprocess.run(
            [
                generator_path,
                "generate",
                "--meta",
                "none",
                "--path",
                f.name,
                f"--custom-template-path={template_path}",
                "--config",
                config_path,
                "--overwrite",
            ],
            cwd=rest_api_path,
            check=True,
        )
    subprocess.run(["git", "add", output_path], check=True, cwd=rest_api_path)
    p = subprocess.run(
        ["pre-commit", "run", "-a"], cwd=rest_api_path, check=False
    )
    if p.returncode:
        subprocess.run(
            ["git", "add", "--update"], check=True, cwd=rest_api_path
        )
        subprocess.run(
            ["pre-commit", "run", "-a"], cwd=rest_api_path, check=False
        )


if __name__ == "__main__":
    spec = get_and_fix_spec(sys.argv[1], sys.argv[2])
    generate_client(spec)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import json
import ssl
from unittest import mock

import pytest
import pytest_httpx

import polarion_rest_api_client as polarion_api
from tests.conftest import TEST_PROJECT_RESPONSE_JSON


@mock.patch("httpx.Client")
@pytest.mark.parametrize("verify_ssl", [False, True])
def test_verify_ssl_boolean(mock_httpx_client, verify_ssl: bool | None):
    polarion_client = polarion_api.PolarionClient(
        polarion_api_endpoint="https://example.com",
        polarion_access_token="fake_token",
        verify_ssl=verify_ssl,
    )

    polarion_client.client.get_httpx_client()

    _, kwargs = mock_httpx_client.call_args
    assert kwargs["verify"] is verify_ssl


@mock.patch("httpx.Client")
def test_default_verify_ssl(mock_httpx_client):
    polarion_client = polarion_api.PolarionClient(
        polarion_api_endpoint="https://example.com",
        polarion_access_token="fake_token",
    )

    polarion_client.client.get_httpx_client()

    _, kwargs = mock_httpx_client.call_args
    assert isinstance(kwargs["verify"], ssl.SSLContext)
    assert kwargs["verify"].protocol == ssl.PROTOCOL_TLS_CLIENT


@mock.patch("httpx.Client")
def test_custom_ssl_context(mock_httpx_client):
    custom_ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    polarion_client = polarion_api.PolarionClient(
        polarion_api_endpoint="https://example.com",
        polarion_access_token="fake_token",
        verify_ssl=custom_ssl_context,
    )

    polarion_client.client.get_httpx_client()

    _, kwargs = mock_httpx_client.call_args
    assert kwargs["verify"] is custom_ssl_context


def test_api_authentication(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_PROJECT_RESPONSE_JSON, encoding="utf8") as f:
        httpx_mock.add_response(
            match_headers={"Authorization": "Bearer PAT123"},
            json=json.load(f),
        )

    assert client.exists()


def test_check_existing_project(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_PROJECT_RESPONSE_JSON, encoding="utf8") as f:
        httpx_mock.add_response(json=json.load(f))

    assert client.exists()


def test_check_non_existing_project(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    httpx_mock.add_response(status_code=404, json={})

    assert not client.exists()
